#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

"""Tests for saml2.md"""

__author__ = 'roland.hedberg@umu.se (Roland Hedberg)'

import unittest

try:
    from xml.etree import ElementTree
except ImportError:
    from elementtree import ElementTree

import saml2
from saml2 import xmldsig as ds
from saml2 import saml
from saml2 import samlp
from saml2 import md
from saml2.extension import idpdisc
from saml2.extension import shibmd

from saml2 import extension_element_to_element
import md_data, ds_data


class TestEndpointType:
    def setup_class(self):
        self.endpoint = md.EndpointType_()

    def testAccessors(self):
        """Test for EndpointType accessors"""
        self.endpoint.binding = saml2.BINDING_HTTP_POST
        self.endpoint.location = "http://www.example.com/endpoint"
        self.endpoint.response_location = "http://www.example.com/response"
        print(self.endpoint.__class__.c_attributes.items())
        new_endpoint = md.endpoint_type__from_string(self.endpoint.to_string())
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"

    def testUsingTestData(self):
        """Test for endpoint_type_from_string() using test data."""
        new_endpoint = md.endpoint_type__from_string(md_data.TEST_ENDPOINT)
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"


class TestIndexedEndpointType:
    def setup_class(self):
        self.i_e = md.IndexedEndpointType_()

    def testAccessors(self):
        """Test for IndexedEndpointType accessors"""
        self.i_e.binding = saml2.BINDING_HTTP_POST
        self.i_e.location = "http://www.example.com/endpoint"
        self.i_e.response_location = "http://www.example.com/response"
        self.i_e.index = "1"
        self.i_e.is_default = "false"
        new_i_e = md.indexed_endpoint_type__from_string(self.i_e.to_string())
        assert new_i_e.binding == saml2.BINDING_HTTP_POST
        assert new_i_e.location == "http://www.example.com/endpoint"
        assert new_i_e.response_location == "http://www.example.com/response"
        assert new_i_e.index == "1"
        assert new_i_e.is_default == "false"

    def testUsingTestData(self):
        """Test for indexed_endpoint_type_from_string() using test data."""
        new_i_e = md.indexed_endpoint_type__from_string(
            md_data.TEST_INDEXED_ENDPOINT)
        assert new_i_e.binding == saml2.BINDING_HTTP_POST
        assert new_i_e.location == "http://www.example.com/endpoint"
        assert new_i_e.response_location == "http://www.example.com/response"
        assert new_i_e.index == "1"
        assert new_i_e.is_default == "false"


class TestExtensions:
    def setup_class(self):
        self.extensions = md.Extensions()

    def testAccessors(self):
        """Test for Extensions accessors"""
        self.extensions.extension_elements.append(
            saml2.extension_element_from_string(
                """<?xml version='1.0' encoding='UTF-8'?>
                <hoge>fuga</hoge>
                """))
        new_extensions = md.extensions_from_string(self.extensions.to_string())
        assert new_extensions.extension_elements[0].tag == "hoge"
        assert new_extensions.extension_elements[0].text.strip() == "fuga"


class TestOrganizationName:
    def setup_class(self):
        self.organization_name = md.OrganizationName()

    def testAccessors(self):
        """Test for OrganizationName accessors"""
        self.organization_name.lang = "en"
        self.organization_name.text = "SIOS Technology, Inc."
        new_organization_name = md.organization_name_from_string(
            self.organization_name.to_string())
        assert new_organization_name.lang == "en"
        assert new_organization_name.text.strip() == "SIOS Technology, Inc."

    def testUsingTestData(self):
        """Test for organization_name_from_string() using test data."""
        new_organization_name = md.organization_name_from_string(
            md_data.TEST_ORGANIZATION_NAME)
        print(new_organization_name.keyswv())
        assert new_organization_name.lang == "se"
        assert new_organization_name.text.strip() == "Catalogix"


class TestOrganizationDisplayName:
    def setup_class(self):
        self.od_name = md.OrganizationDisplayName()

    def testAccessors(self):
        """Test for OrganizationDisplayName accessors"""
        self.od_name.lang = "en"
        self.od_name.text = "SIOS"
        new_od_name = md.organization_display_name_from_string(
            self.od_name.to_string())
        assert new_od_name.lang == "en"
        assert new_od_name.text.strip() == "SIOS"

    def testUsingTestData(self):
        """Test for organization_display_name_from_string() using test data."""
        new_od_name = md.organization_display_name_from_string(
            md_data.TEST_ORGANIZATION_DISPLAY_NAME)
        assert new_od_name.lang == "se"
        assert new_od_name.text.strip() == "Catalogix"


class TestOrganizationURL:
    def setup_class(self):
        self.organization_url = md.OrganizationURL()

    def testAccessors(self):
        """Test for OrganizationURL accessors"""
        self.organization_url.lang = "ja"
        self.organization_url.text = "http://www.example.com/"
        print(self.organization_url.to_string())
        new_organization_url = md.organization_url_from_string(
            self.organization_url.to_string())
        assert new_organization_url.lang == "ja"
        assert new_organization_url.text.strip() == "http://www.example.com/"

    def testUsingTestData(self):
        """Test for organization_url_from_string() using test data."""
        new_organization_url = md.organization_url_from_string(
            md_data.TEST_ORGANIZATION_URL)
        assert new_organization_url.lang == "no"
        assert new_organization_url.text.strip() == "http://www.example.com/"


class TestOrganization:
    def setup_class(self):
        self.organization = md.Organization()

    def testAccessors(self):
        """Test for Organization accessors"""
        self.organization.extensions = md.Extensions()
        self.organization.organization_name.append(
            md.organization_name_from_string(md_data.TEST_ORGANIZATION_NAME))
        self.organization.organization_display_name.append(
            md.organization_display_name_from_string(
                md_data.TEST_ORGANIZATION_DISPLAY_NAME))
        self.organization.organization_url.append(
            md.organization_url_from_string(md_data.TEST_ORGANIZATION_URL))
        new_organization = md.organization_from_string(
            self.organization.to_string())
        assert isinstance(new_organization.extensions, md.Extensions)
        assert isinstance(new_organization.organization_name[0],
                          md.OrganizationName)
        assert isinstance(new_organization.organization_display_name[0],
                          md.OrganizationDisplayName)
        assert isinstance(new_organization.organization_url[0],
                          md.OrganizationURL)
        assert new_organization.organization_name[0].text.strip() == "Catalogix"
        assert new_organization.organization_name[0].lang == "se"
        assert new_organization.organization_display_name[
                   0].text.strip() == "Catalogix"
        assert new_organization.organization_display_name[0].lang == "se"
        assert new_organization.organization_url[
                   0].text.strip() == "http://www.example.com/"
        assert new_organization.organization_url[0].lang == "no"

    def testUsingTestData(self):
        """Test for organization_from_string() using test data."""
        new_organization = md.organization_from_string(
            md_data.TEST_ORGANIZATION)
        assert isinstance(new_organization.extensions, md.Extensions)
        assert isinstance(new_organization.organization_name[0],
                          md.OrganizationName)
        assert isinstance(new_organization.organization_display_name[0],
                          md.OrganizationDisplayName)
        assert isinstance(new_organization.organization_url[0],
                          md.OrganizationURL)
        assert new_organization.organization_name[
                   0].text.strip() == "Catalogix AB"
        assert new_organization.organization_name[0].lang == "se"
        assert new_organization.organization_display_name[
                   0].text.strip() == "Catalogix AS"
        assert new_organization.organization_display_name[0].lang == "no"
        assert new_organization.organization_url[
                   0].text.strip() == "http://www.example.com/"
        assert new_organization.organization_url[0].lang == "en"


class TestContactPerson:
    def setup_class(self):
        self.contact_person = md.ContactPerson()

    def testAccessors(self):
        """Test for ContactPerson accessors"""
        self.contact_person.contact_type = "technical"
        self.contact_person.extensions = md.Extensions()
        self.contact_person.company = md.Company(text="SIOS Technology, Inc.")
        self.contact_person.given_name = md.GivenName(text="Takashi")
        self.contact_person.sur_name = md.SurName(text="Matsuo")
        self.contact_person.email_address.append(
            md.EmailAddress(text="tmatsuo@example.com"))
        self.contact_person.email_address.append(
            md.EmailAddress(text="tmatsuo@shehas.net"))
        self.contact_person.telephone_number.append(
            md.TelephoneNumber(text="00-0000-0000"))
        new_contact_person = md.contact_person_from_string(
            self.contact_person.to_string())
        assert new_contact_person.contact_type == "technical"
        assert isinstance(new_contact_person.extensions, md.Extensions)
        assert new_contact_person.company.text.strip() == "SIOS Technology, " \
                                                          "Inc."
        assert new_contact_person.given_name.text.strip() == "Takashi"
        assert new_contact_person.sur_name.text.strip() == "Matsuo"
        assert new_contact_person.email_address[
                   0].text.strip() == "tmatsuo@example.com"
        assert new_contact_person.email_address[
                   1].text.strip() == "tmatsuo@shehas.net"
        assert new_contact_person.telephone_number[
                   0].text.strip() == "00-0000-0000"

    def testUsingTestData(self):
        """Test for contact_person_from_string() using test data."""
        new_contact_person = md.contact_person_from_string(
            md_data.TEST_CONTACT_PERSON)
        assert new_contact_person.contact_type == "technical"
        assert isinstance(new_contact_person.extensions, md.Extensions)
        assert new_contact_person.company.text.strip() == "SIOS Technology, " \
                                                          "Inc."
        assert new_contact_person.given_name.text.strip() == "Takashi"
        assert new_contact_person.sur_name.text.strip() == "Matsuo"
        assert new_contact_person.email_address[
                   0].text.strip() == "tmatsuo@example.com"
        assert new_contact_person.email_address[
                   1].text.strip() == "tmatsuo@shehas.net"
        assert new_contact_person.telephone_number[
                   0].text.strip() == "00-0000-0000"


class TestAdditionalMetadataLocation:
    def setup_class(self):
        self.additional_metadata_location = md.AdditionalMetadataLocation()

    def testAccessors(self):
        """Test for AdditionalMetadataLocation accessors"""
        self.additional_metadata_location.namespace = (
            "http://www.example.com/namespace")
        self.additional_metadata_location.text = (
            "http://www.example.com/AdditionalMetadataLocation")
        new_additional_metadata_location = \
          md.additional_metadata_location_from_string(
            self.additional_metadata_location.to_string())
        assert new_additional_metadata_location.namespace == \
               "http://www.example.com/namespace"
        assert new_additional_metadata_location.text.strip() == \
               "http://www.example.com/AdditionalMetadataLocation"

    def testUsingTestData(self):
        """Test for additional_metadata_location_from_string() using test
        data."""
        new_additional_metadata_location = \
          md.additional_metadata_location_from_string(
            md_data.TEST_ADDITIONAL_METADATA_LOCATION)
        assert new_additional_metadata_location.namespace == \
               "http://www.example.com/namespace"
        assert new_additional_metadata_location.text.strip() == \
               "http://www.example.com/AdditionalMetadataLocation"


# class TestKeySize:
# 
#   def setup_class(self):
#     self.key_size = md.KeySize()
# 
#   def testAccessors(self):
#     """Test for KeySize accessors"""
#     self.key_size.text = "128"
#     new_key_size = md.key_size_from_string(self.key_size.to_string())
#     assert new_key_size.text.strip() == "128"
# 
#   def testUsingTestData(self):
#     """Test for key_size_from_string() using test data."""
#     new_key_size = md.key_size_from_string(md_data.TEST_KEY_SIZE)
#     assert new_key_size.text.strip() == "128"


# class TestOAEPparams:
# 
#   def setup_class(self):
#     self.oaep_params = md.OAEPparams()
# 
#   def testAccessors(self):
#     """Test for OAEPparams accessors"""
#     self.oaep_params.text = "9lWu3Q=="
#     new_oaep_params = md.oae_pparams_from_string(self.oaep_params.to_string())
#     assert new_oaep_params.text.strip() == "9lWu3Q=="
# 
#   def testUsingTestData(self):
#     """Test for oae_pparams_from_string() using test data."""
#     new_oaep_params = md.oae_pparams_from_string(md_data.TEST_OAEP_PARAMS)
#     assert new_oaep_params.text.strip() == "9lWu3Q=="


class TestEncryptionMethod:
    def setup_class(self):
        self.encryption_method = md.EncryptionMethod()

    def testAccessors(self):
        """Test for EncryptionMethod accessors"""
        self.encryption_method.algorithm = (
            "http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p")
        new_encryption_method = md.encryption_method_from_string(
            self.encryption_method.to_string())
        assert new_encryption_method.algorithm == \
               "http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p"

    def testUsingTestData(self):
        """Test for encryption_method_from_string() using test data."""
        new_encryption_method = md.encryption_method_from_string(
            md_data.TEST_ENCRYPTION_METHOD)
        assert new_encryption_method.algorithm == \
               "http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p"
        assert new_encryption_method.oae_pparams.text.strip() == "9lWu3Q=="


class TestKeyDescriptor:
    def setup_class(self):
        self.key_descriptor = md.KeyDescriptor()

    def testAccessors(self):
        """Test for KeyDescriptor accessors"""

        self.key_descriptor.use = "signing"
        self.key_descriptor.key_info = ds.key_info_from_string(
            ds_data.TEST_KEY_INFO)
        self.key_descriptor.encryption_method.append(
            md.encryption_method_from_string(
                md_data.TEST_ENCRYPTION_METHOD))
        new_key_descriptor = md.key_descriptor_from_string(
            self.key_descriptor.to_string())
        assert new_key_descriptor.use == "signing"
        assert isinstance(new_key_descriptor.key_info, ds.KeyInfo)
        assert isinstance(new_key_descriptor.encryption_method[0],
                          md.EncryptionMethod)

    def testUsingTestData(self):
        """Test for key_descriptor_from_string() using test data."""
        new_key_descriptor = md.key_descriptor_from_string(
            md_data.TEST_KEY_DESCRIPTOR)
        assert new_key_descriptor.use == "signing"
        assert isinstance(new_key_descriptor.key_info, ds.KeyInfo)
        assert isinstance(new_key_descriptor.encryption_method[0],
                          md.EncryptionMethod)


class TestRoleDescriptor:
    def setup_class(self):
        self.role_descriptor = md.RoleDescriptor()

    def testAccessors(self):
        """Test for RoleDescriptor accessors"""
        self.role_descriptor.id = "ID"
        self.role_descriptor.valid_until = "2008-09-14T01:05:02Z"
        self.role_descriptor.cache_duration = "10:00:00:00"
        self.role_descriptor.protocol_support_enumeration = samlp.NAMESPACE
        self.role_descriptor.error_url = "http://www.example.com/errorURL"
        self.role_descriptor.signature = ds.Signature()
        self.role_descriptor.extensions = md.Extensions()
        self.role_descriptor.key_descriptor.append(
            md.key_descriptor_from_string(
                md_data.TEST_KEY_DESCRIPTOR))
        self.role_descriptor.organization = md.Organization()
        self.role_descriptor.contact_person.append(md.ContactPerson())

        new_role_descriptor = md.role_descriptor_from_string(
            self.role_descriptor.to_string())
        assert new_role_descriptor.id == "ID"
        assert new_role_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_role_descriptor.cache_duration == "10:00:00:00"
        assert new_role_descriptor.protocol_support_enumeration == \
               samlp.NAMESPACE
        assert new_role_descriptor.error_url == \
               "http://www.example.com/errorURL"
        assert isinstance(new_role_descriptor.signature, ds.Signature)
        assert isinstance(new_role_descriptor.extensions, md.Extensions)
        assert isinstance(new_role_descriptor.key_descriptor[0],
                          md.KeyDescriptor)
        assert isinstance(new_role_descriptor.organization, md.Organization)
        assert isinstance(new_role_descriptor.contact_person[0],
                          md.ContactPerson)

    def testUsingTestData(self):
        """Test for role_descriptor_from_string() using test data."""
        new_role_descriptor = md.role_descriptor_from_string(
            md_data.TEST_ROLE_DESCRIPTOR)
        assert new_role_descriptor.id == "ID"
        assert new_role_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_role_descriptor.cache_duration == "10:00:00:00"
        assert new_role_descriptor.protocol_support_enumeration == \
               samlp.NAMESPACE
        assert new_role_descriptor.error_url == \
               "http://www.example.com/errorURL"
        assert isinstance(new_role_descriptor.signature, ds.Signature)
        assert isinstance(new_role_descriptor.extensions, md.Extensions)
        assert isinstance(new_role_descriptor.key_descriptor[0],
                          md.KeyDescriptor)
        assert isinstance(new_role_descriptor.organization, md.Organization)
        assert isinstance(new_role_descriptor.contact_person[0],
                          md.ContactPerson)


# class TestSSODescriptor:
#   def setup_class(self):
#     self.sso_descriptor = md.SSODescriptorType_()
# 
#   def testAccessors(self):
#     """Test for SSODescriptorType accessors"""
#     self.sso_descriptor.id = "ID"
#     self.sso_descriptor.valid_until = "2008-09-14T01:05:02Z"
#     self.sso_descriptor.cache_duration = "10:00:00:00"
#     self.sso_descriptor.protocol_support_enumeration = samlp.NAMESPACE
#     self.sso_descriptor.error_url = "http://www.example.com/errorURL"
#     self.sso_descriptor.signature = ds.Signature()
#     self.sso_descriptor.extensions = md.Extensions()
#     self.sso_descriptor.key_descriptor.append(md.key_descriptor_from_string(
#       md_data.TEST_KEY_DESCRIPTOR))
#     self.sso_descriptor.organization = md.Organization()
#     self.sso_descriptor.contact_person.append(md.ContactPerson())
#     self.sso_descriptor.artifact_resolution_service.append(
#       md.ArtifactResolutionService())
#     self.sso_descriptor.single_logout_service.append(
#       md.SingleLogoutService())
#     self.sso_descriptor.manage_name_id_service.append(
#       md.ManageNameIDService())
#     self.sso_descriptor.name_id_format.append(
#       md.NameIDFormat())
# 
#     new_sso_descriptor = md.sso_descriptor_type__from_string(
#       self.sso_descriptor.to_string())
#     assert new_sso_descriptor.id == "ID"
#     assert new_sso_descriptor.valid_until == "2008-09-14T01:05:02Z"
#     assert new_sso_descriptor.cache_duration == "10:00:00:00"
#     assert new_sso_descriptor.protocol_support_enumeration == samlp.NAMESPACE
#     assert new_sso_descriptor.error_url == "http://www.example.com/errorURL"
#     assert isinstance(new_sso_descriptor.signature, ds.Signature)
#     assert isinstance(new_sso_descriptor.extensions, md.Extensions)
#     assert isinstance(new_sso_descriptor.key_descriptor[0],
#                             md.KeyDescriptor)
#     assert isinstance(new_sso_descriptor.organization, md.Organization)
#     assert isinstance(new_sso_descriptor.contact_person[0],
#                             md.ContactPerson)
#     assert isinstance(new_sso_descriptor.artifact_resolution_service[0],
#                             md.ArtifactResolutionService)
#     assert isinstance(new_sso_descriptor.single_logout_service[0],
#                             md.SingleLogoutService)
#     assert isinstance(new_sso_descriptor.manage_name_id_service[0],
#                             md.ManageNameIDService)
#     assert isinstance(new_sso_descriptor.name_id_format[0],
#                             md.NameIDFormat)
# 

class TestArtifactResolutionService:
    def setup_class(self):
        self.i_e = md.ArtifactResolutionService()

    def testAccessors(self):
        """Test for ArtifactResolutionService accessors"""
        self.i_e.binding = saml2.BINDING_HTTP_POST
        self.i_e.location = "http://www.example.com/endpoint"
        self.i_e.response_location = "http://www.example.com/response"
        self.i_e.index = "1"
        self.i_e.is_default = "false"
        new_i_e = md.artifact_resolution_service_from_string(
            self.i_e.to_string())
        assert new_i_e.binding == saml2.BINDING_HTTP_POST
        assert new_i_e.location == "http://www.example.com/endpoint"
        assert new_i_e.response_location == "http://www.example.com/response"
        assert new_i_e.index == "1"
        assert new_i_e.is_default == "false"

    def testUsingTestData(self):
        """Test for artifact_resolution_service_from_string() using test
        data."""
        new_i_e = md.artifact_resolution_service_from_string(
            md_data.TEST_ARTIFACT_RESOLUTION_SERVICE)
        assert new_i_e.binding == saml2.BINDING_HTTP_POST
        assert new_i_e.location == "http://www.example.com/endpoint"
        assert new_i_e.response_location == "http://www.example.com/response"
        assert new_i_e.index == "1"
        assert new_i_e.is_default == "false"


class TestSingleLogout:
    def setup_class(self):
        self.endpoint = md.SingleLogoutService()

    def testAccessors(self):
        """Test for SingleLogoutService accessors"""
        self.endpoint.binding = saml2.BINDING_HTTP_POST
        self.endpoint.location = "http://www.example.com/endpoint"
        self.endpoint.response_location = "http://www.example.com/response"
        new_endpoint = md.single_logout_service_from_string(
            self.endpoint.to_string())
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"

    def testUsingTestData(self):
        """Test for single_logout_service_from_string() using test data."""
        new_endpoint = md.single_logout_service_from_string(
            md_data.TEST_SINGLE_LOGOUT_SERVICE)
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"


class TestManageNameIDService:
    def setup_class(self):
        self.endpoint = md.ManageNameIDService()

    def testAccessors(self):
        """Test for ManageNameIDService accessors"""
        self.endpoint.binding = saml2.BINDING_HTTP_POST
        self.endpoint.location = "http://www.example.com/endpoint"
        self.endpoint.response_location = "http://www.example.com/response"
        new_endpoint = md.manage_name_id_service_from_string(
            self.endpoint.to_string())
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"

    def testUsingTestData(self):
        """Test for manage_name_id_service_from_string() using test data."""
        new_endpoint = md.manage_name_id_service_from_string(
            md_data.TEST_MANAGE_NAMEID_SERVICE)
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"


class TestNameIDFormat:
    def setup_class(self):
        self.name_id_format = md.NameIDFormat()

    def testAccessors(self):
        """Test for NameIDFormat accessors"""
        self.name_id_format.text = saml.NAMEID_FORMAT_EMAILADDRESS
        new_name_id_format = md.name_id_format_from_string(
            self.name_id_format.to_string())
        assert new_name_id_format.text.strip() == \
               saml.NAMEID_FORMAT_EMAILADDRESS

    def testUsingTestData(self):
        """Test for name_id_format_from_string() using test data."""
        new_name_id_format = md.name_id_format_from_string(
            md_data.TEST_NAME_ID_FORMAT)
        assert new_name_id_format.text.strip() == \
               saml.NAMEID_FORMAT_EMAILADDRESS


class TestSingleSignOnService:
    def setup_class(self):
        self.endpoint = md.SingleSignOnService()

    def testAccessors(self):
        """Test for SingelSignOnService accessors"""
        self.endpoint.binding = saml2.BINDING_HTTP_POST
        self.endpoint.location = "http://www.example.com/endpoint"
        self.endpoint.response_location = "http://www.example.com/response"
        new_endpoint = md.single_sign_on_service_from_string(
            self.endpoint.to_string())
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"

    def testUsingTestData(self):
        """Test for SingelSignOn_service_from_string() using test data."""
        new_endpoint = md.single_sign_on_service_from_string(
            md_data.TEST_SINGLE_SIGN_ON_SERVICE)
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"


class TestNameIDMappingService:
    def setup_class(self):
        self.endpoint = md.NameIDMappingService()

    def testAccessors(self):
        """Test for NameIDMappingService accessors"""
        self.endpoint.binding = saml2.BINDING_HTTP_POST
        self.endpoint.location = "http://www.example.com/endpoint"
        self.endpoint.response_location = "http://www.example.com/response"
        new_endpoint = md.name_id_mapping_service_from_string(
            self.endpoint.to_string())
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"

    def testUsingTestData(self):
        """Test for name_id_mapping_service_from_string() using test data."""
        new_endpoint = md.name_id_mapping_service_from_string(
            md_data.TEST_NAME_ID_MAPPING_SERVICE)
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"


class TestAssertionIDRequestService:
    def setup_class(self):
        self.endpoint = md.AssertionIDRequestService()

    def testAccessors(self):
        """Test for AssertionIDRequestService accessors"""
        self.endpoint.binding = saml2.BINDING_HTTP_POST
        self.endpoint.location = "http://www.example.com/endpoint"
        self.endpoint.response_location = "http://www.example.com/response"
        new_endpoint = md.assertion_id_request_service_from_string(
            self.endpoint.to_string())
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"

    def testUsingTestData(self):
        """Test for assertion_id_request_service_from_string() using test
        data."""
        new_endpoint = md.assertion_id_request_service_from_string(
            md_data.TEST_ASSERTION_ID_REQUEST_SERVICE)
        assert new_endpoint.binding == saml2.BINDING_HTTP_POST
        assert new_endpoint.location == "http://www.example.com/endpoint"
        assert new_endpoint.response_location == \
               "http://www.example.com/response"


class TestAttributeProfile:
    def setup_class(self):
        self.attribute_profile = md.AttributeProfile()

    def testAccessors(self):
        """Test for AttributeProfile accessors"""
        self.attribute_profile.text = saml.PROFILE_ATTRIBUTE_BASIC
        new_attribute_profile = md.attribute_profile_from_string(
            self.attribute_profile.to_string())
        assert new_attribute_profile.text.strip() == \
               saml.PROFILE_ATTRIBUTE_BASIC

    def testUsingTestData(self):
        """Test for name_id_format_from_string() using test data."""
        new_attribute_profile = md.attribute_profile_from_string(
            md_data.TEST_ATTRIBUTE_PROFILE)
        assert new_attribute_profile.text.strip() == \
               saml.PROFILE_ATTRIBUTE_BASIC


class TestIDPSSODescriptor:
    def setup_class(self):
        self.idp_sso_descriptor = md.IDPSSODescriptor()

    def testAccessors(self):
        """Test for IDPSSODescriptor accessors"""
        self.idp_sso_descriptor.id = "ID"
        self.idp_sso_descriptor.valid_until = "2008-09-14T01:05:02Z"
        self.idp_sso_descriptor.cache_duration = "10:00:00:00"
        self.idp_sso_descriptor.protocol_support_enumeration = \
            samlp.NAMESPACE
        self.idp_sso_descriptor.error_url = "http://www.example.com/errorURL"
        self.idp_sso_descriptor.signature = ds.Signature()
        self.idp_sso_descriptor.extensions = md.Extensions()
        self.idp_sso_descriptor.key_descriptor.append(
            md.key_descriptor_from_string(
                md_data.TEST_KEY_DESCRIPTOR))
        self.idp_sso_descriptor.organization = md.Organization()
        self.idp_sso_descriptor.contact_person.append(md.ContactPerson())
        self.idp_sso_descriptor.artifact_resolution_service.append(
            md.ArtifactResolutionService())
        self.idp_sso_descriptor.single_logout_service.append(
            md.SingleLogoutService())
        self.idp_sso_descriptor.manage_name_id_service.append(
            md.ManageNameIDService())
        self.idp_sso_descriptor.name_id_format.append(
            md.NameIDFormat())
        self.idp_sso_descriptor.want_authn_requests_signed = 'true'
        self.idp_sso_descriptor.single_sign_on_service.append(
            md.SingleSignOnService())
        self.idp_sso_descriptor.name_id_mapping_service.append(
            md.NameIDMappingService())
        self.idp_sso_descriptor.assertion_id_request_service.append(
            md.AssertionIDRequestService())
        self.idp_sso_descriptor.attribute_profile.append(
            md.AttributeProfile())
        self.idp_sso_descriptor.attribute.append(saml.Attribute())

        new_idp_sso_descriptor = md.idpsso_descriptor_from_string(
            self.idp_sso_descriptor.to_string())
        assert new_idp_sso_descriptor.id == "ID"
        assert new_idp_sso_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_idp_sso_descriptor.cache_duration == "10:00:00:00"
        assert new_idp_sso_descriptor.protocol_support_enumeration == \
               samlp.NAMESPACE
        assert new_idp_sso_descriptor.error_url == \
               "http://www.example.com/errorURL"
        assert isinstance(new_idp_sso_descriptor.signature, ds.Signature)
        assert isinstance(new_idp_sso_descriptor.extensions, md.Extensions)
        assert isinstance(new_idp_sso_descriptor.key_descriptor[0],
                          md.KeyDescriptor)
        assert isinstance(new_idp_sso_descriptor.organization,
                          md.Organization)
        assert isinstance(new_idp_sso_descriptor.contact_person[0],
                          md.ContactPerson)
        assert isinstance(
            new_idp_sso_descriptor.artifact_resolution_service[0],
            md.ArtifactResolutionService)
        assert isinstance(new_idp_sso_descriptor.single_logout_service[0],
                          md.SingleLogoutService)
        assert isinstance(new_idp_sso_descriptor.manage_name_id_service[0],
                          md.ManageNameIDService)
        assert isinstance(new_idp_sso_descriptor.name_id_format[0],
                          md.NameIDFormat)
        assert new_idp_sso_descriptor.want_authn_requests_signed == "true"
        assert isinstance(new_idp_sso_descriptor.single_sign_on_service[0],
                          md.SingleSignOnService)
        assert isinstance(new_idp_sso_descriptor.name_id_mapping_service[0],
                          md.NameIDMappingService)
        assert isinstance(
            new_idp_sso_descriptor.assertion_id_request_service[0],
            md.AssertionIDRequestService)
        assert isinstance(new_idp_sso_descriptor.attribute_profile[0],
                          md.AttributeProfile)
        assert isinstance(new_idp_sso_descriptor.attribute[0],
                          saml.Attribute)

    def testUsingTestData(self):
        """Test for idpsso_descriptor_from_string() using test data."""
        new_idp_sso_descriptor = md.idpsso_descriptor_from_string(
            md_data.TEST_IDP_SSO_DESCRIPTOR)
        assert new_idp_sso_descriptor.id == "ID"
        assert new_idp_sso_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_idp_sso_descriptor.cache_duration == "10:00:00:00"
        assert new_idp_sso_descriptor.protocol_support_enumeration == \
               samlp.NAMESPACE
        assert new_idp_sso_descriptor.error_url == \
               "http://www.example.com/errorURL"
        assert isinstance(new_idp_sso_descriptor.signature, ds.Signature)
        assert isinstance(new_idp_sso_descriptor.extensions, md.Extensions)
        assert isinstance(new_idp_sso_descriptor.key_descriptor[0],
                          md.KeyDescriptor)
        assert isinstance(new_idp_sso_descriptor.organization,
                          md.Organization)
        assert isinstance(new_idp_sso_descriptor.contact_person[0],
                          md.ContactPerson)
        assert isinstance(
            new_idp_sso_descriptor.artifact_resolution_service[0],
            md.ArtifactResolutionService)
        assert isinstance(new_idp_sso_descriptor.single_logout_service[0],
                          md.SingleLogoutService)
        assert isinstance(new_idp_sso_descriptor.manage_name_id_service[0],
                          md.ManageNameIDService)
        assert isinstance(new_idp_sso_descriptor.name_id_format[0],
                          md.NameIDFormat)
        assert new_idp_sso_descriptor.want_authn_requests_signed == "true"
        assert isinstance(new_idp_sso_descriptor.single_sign_on_service[0],
                          md.SingleSignOnService)
        assert isinstance(new_idp_sso_descriptor.name_id_mapping_service[0],
                          md.NameIDMappingService)
        assert isinstance(
            new_idp_sso_descriptor.assertion_id_request_service[0],
            md.AssertionIDRequestService)
        assert isinstance(new_idp_sso_descriptor.attribute_profile[0],
                          md.AttributeProfile)
        assert isinstance(new_idp_sso_descriptor.attribute[0],
                          saml.Attribute)

    def testUsingScope(self):
        descriptor = md.IDPSSODescriptor()
        scope = shibmd.Scope()
        scope.text = "example.org"
        scope.regexp = "false"
        descriptor.extensions = md.Extensions()
        ext = saml2.element_to_extension_element(scope)
        descriptor.extensions.extension_elements.append(ext)
        exts = descriptor.extensions
        assert len(exts.extension_elements) == 1
        elem = exts.extension_elements[0]
        inst = saml2.extension_element_to_element(elem,
                                                  shibmd.ELEMENT_FROM_STRING,
                                                  namespace=shibmd.NAMESPACE)
        assert isinstance(inst, shibmd.Scope)
        assert inst.text == "example.org"
        assert inst.regexp == "false"


class TestAssertionConsumerService:
    def setup_class(self):
        self.i_e = md.AssertionConsumerService()

    def testAccessors(self):
        """Test for AssertionConsumerService accessors"""
        self.i_e.binding = saml2.BINDING_HTTP_POST
        self.i_e.location = "http://www.example.com/endpoint"
        self.i_e.response_location = "http://www.example.com/response"
        self.i_e.index = "1"
        self.i_e.is_default = "false"
        new_i_e = md.assertion_consumer_service_from_string(
            self.i_e.to_string())
        assert new_i_e.binding == saml2.BINDING_HTTP_POST
        assert new_i_e.location == "http://www.example.com/endpoint"
        assert new_i_e.response_location == "http://www.example.com/response"
        assert new_i_e.index == "1"
        assert new_i_e.is_default == "false"

    def testUsingTestData(self):
        """Test for assertion_consumer_service_from_string() using test data."""
        new_i_e = md.assertion_consumer_service_from_string(
            md_data.TEST_ASSERTION_CONSUMER_SERVICE)
        assert new_i_e.binding == saml2.BINDING_HTTP_POST
        assert new_i_e.location == "http://www.example.com/endpoint"
        assert new_i_e.response_location == "http://www.example.com/response"
        assert new_i_e.index == "1"
        assert new_i_e.is_default == "false"


class TestRequestedAttribute:
    def setup_class(self):
        self.requested_attribute = md.RequestedAttribute()

    def testAccessors(self):
        """Test for RequestedAttribute accessors"""
        assert isinstance(self.requested_attribute, saml.AttributeType_)
        assert isinstance(self.requested_attribute, md.RequestedAttribute)
        assert self.requested_attribute.is_required is None
        self.requested_attribute.is_required = "true"
        new_requested_attribute = md.requested_attribute_from_string(
            self.requested_attribute.to_string())
        assert new_requested_attribute.is_required == "true"
        assert isinstance(new_requested_attribute, saml.AttributeType_)
        assert isinstance(new_requested_attribute, md.RequestedAttribute)

    def testUsingTestData(self):
        """Test for requested_attribute_from_string() using test data."""
        new_requested_attribute = md.requested_attribute_from_string(
            md_data.TEST_REQUESTED_ATTRIBUTE)
        assert new_requested_attribute.is_required == "true"
        assert isinstance(new_requested_attribute, saml.AttributeType_)
        assert isinstance(new_requested_attribute, md.RequestedAttribute)


class TestServiceName:
    def setup_class(self):
        self.service_name = md.ServiceName()

    def testAccessors(self):
        """Test for ServiceName accessors"""
        self.service_name.lang = "en"
        self.service_name.text = "SIOS mail"
        new_service_name = md.service_name_from_string(
            self.service_name.to_string())
        assert new_service_name.lang == "en"
        assert new_service_name.text.strip() == "SIOS mail"

    def testUsingTestData(self):
        """Test for organization_name_from_string() using test data."""
        new_service_name = md.service_name_from_string(
            md_data.TEST_SERVICE_NAME)
        assert new_service_name.lang == "en"
        assert new_service_name.text.strip() == "Catalogix Whois"


class TestServiceDescription:
    def setup_class(self):
        self.service_description = md.ServiceDescription()

    def testAccessors(self):
        """Test for ServiceDescription accessors"""
        self.service_description.lang = "en"
        self.service_description.text = "SIOS mail service"
        new_service_description = md.service_description_from_string(
            self.service_description.to_string())
        assert new_service_description.lang == "en"
        assert new_service_description.text.strip() == "SIOS mail service"

    def testUsingTestData(self):
        """Test for organization_name_from_string() using test data."""
        new_service_description = md.service_description_from_string(
            md_data.TEST_SERVICE_DESCRIPTION)
        assert new_service_description.lang == "en"
        assert new_service_description.text.strip() == "Catalogix Whois Service"


class TestAttributeConsumingService:
    def setup_class(self):
        self.attribute_consuming_service = md.AttributeConsumingService()

    def testAccessors(self):
        """Test for AttributeConsumingService accessors"""
        self.attribute_consuming_service.service_name.append(md.ServiceName())
        self.attribute_consuming_service.service_description.append(
            md.ServiceDescription())
        self.attribute_consuming_service.requested_attribute.append(
            md.RequestedAttribute())
        self.attribute_consuming_service.index = "1"
        self.attribute_consuming_service.is_default = "true"

        new_attribute_consuming_service = \
          md.attribute_consuming_service_from_string(
            self.attribute_consuming_service.to_string())
        assert new_attribute_consuming_service.index == "1"
        assert new_attribute_consuming_service.is_default == "true"
        assert isinstance(new_attribute_consuming_service.service_name[0],
                          md.ServiceName)
        assert isinstance(
            new_attribute_consuming_service.service_description[0],
            md.ServiceDescription)
        assert isinstance(
            new_attribute_consuming_service.requested_attribute[0],
            md.RequestedAttribute)

    def testUsingTestData(self):
        """Test for attribute_consuming_service_from_string() using test
        data."""
        new_attribute_consuming_service = \
          md.attribute_consuming_service_from_string(
            md_data.TEST_ATTRIBUTE_CONSUMING_SERVICE)
        assert new_attribute_consuming_service.index == "1"
        assert new_attribute_consuming_service.is_default == "true"
        assert isinstance(new_attribute_consuming_service.service_name[0],
                          md.ServiceName)
        assert isinstance(
            new_attribute_consuming_service.service_description[0],
            md.ServiceDescription)
        assert isinstance(
            new_attribute_consuming_service.requested_attribute[0],
            md.RequestedAttribute)


class TestSPSSODescriptor:
    def setup_class(self):
        self.sp_sso_descriptor = md.SPSSODescriptor()

    def testAccessors(self):
        """Test for SPSSODescriptor accessors"""
        self.sp_sso_descriptor.id = "ID"
        self.sp_sso_descriptor.valid_until = "2008-09-14T01:05:02Z"
        self.sp_sso_descriptor.cache_duration = "10:00:00:00"
        self.sp_sso_descriptor.protocol_support_enumeration = \
            samlp.NAMESPACE
        self.sp_sso_descriptor.error_url = "http://www.example.com/errorURL"
        self.sp_sso_descriptor.signature = ds.Signature()
        self.sp_sso_descriptor.extensions = md.Extensions()
        self.sp_sso_descriptor.key_descriptor.append(
            md.key_descriptor_from_string(
                md_data.TEST_KEY_DESCRIPTOR))
        self.sp_sso_descriptor.organization = md.Organization()
        self.sp_sso_descriptor.contact_person.append(md.ContactPerson())
        self.sp_sso_descriptor.artifact_resolution_service.append(
            md.ArtifactResolutionService())
        self.sp_sso_descriptor.single_logout_service.append(
            md.SingleLogoutService())
        self.sp_sso_descriptor.manage_name_id_service.append(
            md.ManageNameIDService())
        self.sp_sso_descriptor.name_id_format.append(
            md.NameIDFormat())
        self.sp_sso_descriptor.authn_requests_signed = "true"
        self.sp_sso_descriptor.want_assertions_signed = "true"
        self.sp_sso_descriptor.assertion_consumer_service.append(
            md.AssertionConsumerService())
        self.sp_sso_descriptor.attribute_consuming_service.append(
            md.AttributeConsumingService())

        print(self.sp_sso_descriptor)
        new_sp_sso_descriptor = md.spsso_descriptor_from_string(
            self.sp_sso_descriptor.to_string())
        print(new_sp_sso_descriptor)
        assert new_sp_sso_descriptor.id == "ID"
        assert new_sp_sso_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_sp_sso_descriptor.cache_duration == "10:00:00:00"
        assert new_sp_sso_descriptor.protocol_support_enumeration == \
               samlp.NAMESPACE
        assert new_sp_sso_descriptor.error_url == \
               "http://www.example.com/errorURL"
        assert isinstance(new_sp_sso_descriptor.signature, ds.Signature)
        assert isinstance(new_sp_sso_descriptor.extensions, md.Extensions)
        assert isinstance(new_sp_sso_descriptor.key_descriptor[0],
                          md.KeyDescriptor)
        assert isinstance(new_sp_sso_descriptor.organization,
                          md.Organization)
        assert isinstance(new_sp_sso_descriptor.contact_person[0],
                          md.ContactPerson)
        assert isinstance(
            new_sp_sso_descriptor.artifact_resolution_service[0],
            md.ArtifactResolutionService)
        assert isinstance(new_sp_sso_descriptor.single_logout_service[0],
                          md.SingleLogoutService)
        assert isinstance(new_sp_sso_descriptor.manage_name_id_service[0],
                          md.ManageNameIDService)
        assert isinstance(new_sp_sso_descriptor.name_id_format[0],
                          md.NameIDFormat)
        assert new_sp_sso_descriptor.authn_requests_signed == "true"
        assert new_sp_sso_descriptor.want_assertions_signed == "true"
        assert isinstance(
            new_sp_sso_descriptor.assertion_consumer_service[0],
            md.AssertionConsumerService)
        assert isinstance(
            new_sp_sso_descriptor.attribute_consuming_service[0],
            md.AttributeConsumingService)

    def testUsingTestData(self):
        """Test for spsso_descriptor_from_string() using test data."""
        new_sp_sso_descriptor = md.spsso_descriptor_from_string(
            md_data.TEST_SP_SSO_DESCRIPTOR)
        assert new_sp_sso_descriptor.id == "ID"
        assert new_sp_sso_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_sp_sso_descriptor.cache_duration == "10:00:00:00"
        assert new_sp_sso_descriptor.protocol_support_enumeration == \
               samlp.NAMESPACE
        assert new_sp_sso_descriptor.error_url == \
               "http://www.example.com/errorURL"
        assert isinstance(new_sp_sso_descriptor.signature, ds.Signature)
        assert isinstance(new_sp_sso_descriptor.extensions, md.Extensions)
        print(new_sp_sso_descriptor.extensions.__dict__)
        assert len(new_sp_sso_descriptor.extensions.extension_elements) == 2
        for eelem in new_sp_sso_descriptor.extensions.extension_elements:
            print("EE", eelem.__dict__)
            dp = extension_element_to_element(eelem,
                                              idpdisc.ELEMENT_FROM_STRING,
                                              idpdisc.NAMESPACE)
            print("DP", dp.c_tag, dp.c_namespace, dp.__dict__)
            assert isinstance(dp, idpdisc.DiscoveryResponse)
        assert isinstance(new_sp_sso_descriptor.key_descriptor[0],
                          md.KeyDescriptor)
        assert isinstance(new_sp_sso_descriptor.organization,
                          md.Organization)
        assert isinstance(new_sp_sso_descriptor.contact_person[0],
                          md.ContactPerson)
        assert isinstance(
            new_sp_sso_descriptor.artifact_resolution_service[0],
            md.ArtifactResolutionService)
        assert isinstance(new_sp_sso_descriptor.single_logout_service[0],
                          md.SingleLogoutService)
        assert isinstance(new_sp_sso_descriptor.manage_name_id_service[0],
                          md.ManageNameIDService)
        assert isinstance(new_sp_sso_descriptor.name_id_format[0],
                          md.NameIDFormat)
        assert new_sp_sso_descriptor.authn_requests_signed == "true"
        assert new_sp_sso_descriptor.want_assertions_signed == "true"
        assert isinstance(
            new_sp_sso_descriptor.assertion_consumer_service[0],
            md.AssertionConsumerService)
        assert isinstance(
            new_sp_sso_descriptor.attribute_consuming_service[0],
            md.AttributeConsumingService)


class TestEntityDescriptor:
    def setup_class(self):
        self.entity_descriptor = md.EntityDescriptor()

    def testAccessors(self):
        """Test for RoleDescriptor accessors"""
        self.entity_descriptor.id = "ID"
        self.entity_descriptor.entity_id = "entityID"
        self.entity_descriptor.valid_until = "2008-09-14T01:05:02Z"
        self.entity_descriptor.cache_duration = "10:00:00:00"

        self.entity_descriptor.signature = ds.Signature()
        self.entity_descriptor.extensions = md.Extensions()
        self.entity_descriptor.role_descriptor.append(md.RoleDescriptor())
        self.entity_descriptor.idpsso_descriptor.append(md.IDPSSODescriptor())
        self.entity_descriptor.spsso_descriptor.append(md.SPSSODescriptor())
        self.entity_descriptor.organization = md.Organization()
        self.entity_descriptor.contact_person.append(md.ContactPerson())
        self.entity_descriptor.additional_metadata_location.append(
            md.AdditionalMetadataLocation())

        new_entity_descriptor = md.entity_descriptor_from_string(
            self.entity_descriptor.to_string())
        assert new_entity_descriptor.id == "ID"
        assert new_entity_descriptor.entity_id == "entityID"
        assert new_entity_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_entity_descriptor.cache_duration == "10:00:00:00"
        assert isinstance(new_entity_descriptor.signature, ds.Signature)
        assert isinstance(new_entity_descriptor.extensions, md.Extensions)
        assert isinstance(new_entity_descriptor.role_descriptor[0],
                          md.RoleDescriptor)
        assert isinstance(new_entity_descriptor.idpsso_descriptor[0],
                          md.IDPSSODescriptor)
        assert isinstance(new_entity_descriptor.spsso_descriptor[0],
                          md.SPSSODescriptor)
        assert isinstance(new_entity_descriptor.organization,
                          md.Organization)
        assert isinstance(new_entity_descriptor.contact_person[0],
                          md.ContactPerson)
        assert isinstance(
            new_entity_descriptor.additional_metadata_location[0],
            md.AdditionalMetadataLocation)

    def testUsingTestData(self):
        """Test for entity_descriptor_from_string() using test data."""
        new_entity_descriptor = md.entity_descriptor_from_string(
            md_data.TEST_ENTITY_DESCRIPTOR)
        assert new_entity_descriptor.id == "ID"
        assert new_entity_descriptor.entity_id == "entityID"
        assert new_entity_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_entity_descriptor.cache_duration == "10:00:00:00"
        assert isinstance(new_entity_descriptor.signature, ds.Signature)
        assert isinstance(new_entity_descriptor.extensions, md.Extensions)
        assert isinstance(new_entity_descriptor.role_descriptor[0],
                          md.RoleDescriptor)
        assert isinstance(new_entity_descriptor.idpsso_descriptor[0],
                          md.IDPSSODescriptor)
        assert isinstance(new_entity_descriptor.spsso_descriptor[0],
                          md.SPSSODescriptor)
        assert isinstance(new_entity_descriptor.organization,
                          md.Organization)
        assert isinstance(new_entity_descriptor.contact_person[0],
                          md.ContactPerson)
        assert isinstance(new_entity_descriptor.additional_metadata_location[0],
                          md.AdditionalMetadataLocation)


class TestEntitiesDescriptor:
    def setup_class(self):
        self.entities_descriptor = md.EntitiesDescriptor()

    def testAccessors(self):
        """Test for EntitiesDescriptor accessors"""
        self.entities_descriptor.id = "ID"
        self.entities_descriptor.name = "name"
        self.entities_descriptor.valid_until = "2008-09-14T01:05:02Z"
        self.entities_descriptor.cache_duration = "10:00:00:00"

        self.entities_descriptor.signature = ds.Signature()
        self.entities_descriptor.extensions = md.Extensions()
        self.entities_descriptor.entity_descriptor.append(md.EntityDescriptor())
        self.entities_descriptor.entities_descriptor.append(
            md.EntitiesDescriptor())

        new_entities_descriptor = md.entities_descriptor_from_string(
            self.entities_descriptor.to_string())
        assert new_entities_descriptor.id == "ID"
        assert new_entities_descriptor.name == "name"
        assert new_entities_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_entities_descriptor.cache_duration == "10:00:00:00"
        assert isinstance(new_entities_descriptor.signature, ds.Signature)
        assert isinstance(new_entities_descriptor.extensions, md.Extensions)
        assert isinstance(new_entities_descriptor.entity_descriptor[0],
                          md.EntityDescriptor)
        assert isinstance(new_entities_descriptor.entities_descriptor[0],
                          md.EntitiesDescriptor)

    def testUsingTestData(self):
        """Test for entities_descriptor_from_string() using test data."""
        new_entities_descriptor = md.entities_descriptor_from_string(
            md_data.TEST_ENTITIES_DESCRIPTOR)
        assert new_entities_descriptor.id == "ID"
        assert new_entities_descriptor.name == "name"
        assert new_entities_descriptor.valid_until == "2008-09-14T01:05:02Z"
        assert new_entities_descriptor.cache_duration == "10:00:00:00"
        assert isinstance(new_entities_descriptor.signature, ds.Signature)
        assert isinstance(new_entities_descriptor.extensions, md.Extensions)
        assert isinstance(new_entities_descriptor.entity_descriptor[0],
                          md.EntityDescriptor)
        assert isinstance(new_entities_descriptor.entities_descriptor[0],
                          md.EntitiesDescriptor)


if __name__ == "__main__":
    c = TestIDPSSODescriptor()
    c.setup_class()
    c.testAccessors()


#!/usr/bin/env python
import importlib
import argparse
import base64
import re
import logging
import time
from hashlib import sha1

from urlparse import parse_qs
from Cookie import SimpleCookie
import os
from saml2.authn import is_equal
from saml2.profile import ecp

from saml2 import server
from saml2 import BINDING_HTTP_ARTIFACT
from saml2 import BINDING_URI
from saml2 import BINDING_PAOS
from saml2 import BINDING_SOAP
from saml2 import BINDING_HTTP_REDIRECT
from saml2 import BINDING_HTTP_POST
from saml2 import time_util

from saml2.authn_context import AuthnBroker
from saml2.authn_context import PASSWORD
from saml2.authn_context import UNSPECIFIED
from saml2.authn_context import authn_context_class_ref
from saml2.httputil import Response
from saml2.httputil import NotFound
from saml2.httputil import geturl
from saml2.httputil import get_post
from saml2.httputil import Redirect
from saml2.httputil import Unauthorized
from saml2.httputil import BadRequest
from saml2.httputil import ServiceError
from saml2.ident import Unknown
from saml2.metadata import create_metadata_string
from saml2.s_utils import rndstr
from saml2.s_utils import exception_trace
from saml2.s_utils import UnknownPrincipal
from saml2.s_utils import UnsupportedBinding
from saml2.s_utils import PolicyError
from saml2.sigver import verify_redirect_signature
from saml2.sigver import encrypt_cert_from_item

logger = logging.getLogger("saml2.idp")


class Cache(object):
    def __init__(self):
        self.user2uid = {}
        self.uid2user = {}


def _expiration(timeout, tformat="%a, %d-%b-%Y %H:%M:%S GMT"):
    """

    :param timeout:
    :param tformat:
    :return:
    """
    if timeout == "now":
        return time_util.instant(tformat)
    elif timeout == "dawn":
        return time.strftime(tformat, time.gmtime(0))
    else:
        # validity time should match lifetime of assertions
        return time_util.in_a_while(minutes=timeout, format=tformat)


def get_eptid(idp, req_info, session):
    return idp.eptid.get(idp.config.entityid,
                         req_info.sender(), session["permanent_id"],
                         session["authn_auth"])


# -----------------------------------------------------------------------------


def dict2list_of_tuples(d):
    return [(k, v) for k, v in d.items()]


# -----------------------------------------------------------------------------


class Service(object):
    def __init__(self, environ, start_response, user=None):
        self.environ = environ
        logger.debug("ENVIRON: %s", environ)
        self.start_response = start_response
        self.user = user

    def unpack_redirect(self):
        if "QUERY_STRING" in self.environ:
            _qs = self.environ["QUERY_STRING"]
            return dict([(k, v[0]) for k, v in parse_qs(_qs).items()])
        else:
            return None

    def unpack_post(self):
        _dict = parse_qs(get_post(self.environ))
        logger.debug("unpack_post:: %s", _dict)
        try:
            return dict([(k, v[0]) for k, v in _dict.items()])
        except Exception:
            return None

    def unpack_soap(self):
        try:
            query = get_post(self.environ)
            return {"SAMLRequest": query, "RelayState": ""}
        except Exception:
            return None

    def unpack_either(self):
        if self.environ["REQUEST_METHOD"] == "GET":
            _dict = self.unpack_redirect()
        elif self.environ["REQUEST_METHOD"] == "POST":
            _dict = self.unpack_post()
        else:
            _dict = None
        logger.debug("_dict: %s", _dict)
        return _dict

    def operation(self, saml_msg, binding):
        logger.debug("_operation: %s", saml_msg)
        if not saml_msg or not 'SAMLRequest' in saml_msg:
            resp = BadRequest('Error parsing request or no request')
            return resp(self.environ, self.start_response)
        else:
            try:
                _encrypt_cert = encrypt_cert_from_item(
                    saml_msg["req_info"].message)
                return self.do(saml_msg["SAMLRequest"], binding,
                               saml_msg["RelayState"],
                               encrypt_cert=_encrypt_cert)
            except KeyError:
                # Can live with no relay state
                return self.do(saml_msg["SAMLRequest"], binding)

    def artifact_operation(self, saml_msg):
        if not saml_msg:
            resp = BadRequest("Missing query")
            return resp(self.environ, self.start_response)
        else:
            # exchange artifact for request
            request = IDP.artifact2message(saml_msg["SAMLart"], "spsso")
            try:
                return self.do(request, BINDING_HTTP_ARTIFACT,
                               saml_msg["RelayState"])
            except KeyError:
                return self.do(request, BINDING_HTTP_ARTIFACT)

    def response(self, binding, http_args):
        if binding == BINDING_HTTP_ARTIFACT:
            resp = Redirect()
        else:
            resp = Response(http_args["data"], headers=http_args["headers"])
        return resp(self.environ, self.start_response)

    def do(self, query, binding, relay_state="", encrypt_cert=None):
        pass

    def redirect(self):
        """ Expects a HTTP-redirect request """

        _dict = self.unpack_redirect()
        return self.operation(_dict, BINDING_HTTP_REDIRECT)

    def post(self):
        """ Expects a HTTP-POST request """

        _dict = self.unpack_post()
        return self.operation(_dict, BINDING_HTTP_POST)

    def artifact(self):
        # Can be either by HTTP_Redirect or HTTP_POST
        _dict = self.unpack_either()
        return self.artifact_operation(_dict)

    def soap(self):
        """
        Single log out using HTTP_SOAP binding
        """
        logger.debug("- SOAP -")
        _dict = self.unpack_soap()
        logger.debug("_dict: %s", _dict)
        return self.operation(_dict, BINDING_SOAP)

    def uri(self):
        _dict = self.unpack_either()
        return self.operation(_dict, BINDING_SOAP)

    def not_authn(self, key, requested_authn_context):
        ruri = geturl(self.environ, query=False)
        return do_authentication(self.environ, self.start_response,
                                 authn_context=requested_authn_context,
                                 key=key, redirect_uri=ruri)


# -----------------------------------------------------------------------------

REPOZE_ID_EQUIVALENT = "uid"
FORM_SPEC = """<form name="myform" method="post" action="%s">
   <input type="hidden" name="SAMLResponse" value="%s" />
   <input type="hidden" name="RelayState" value="%s" />
</form>"""

# -----------------------------------------------------------------------------
# === Single log in ====
# -----------------------------------------------------------------------------


class AuthenticationNeeded(Exception):
    def __init__(self, authn_context=None, *args, **kwargs):
        Exception.__init__(*args, **kwargs)
        self.authn_context = authn_context


class SSO(Service):
    def __init__(self, environ, start_response, user=None):
        Service.__init__(self, environ, start_response, user)
        self.binding = ""
        self.response_bindings = None
        self.resp_args = {}
        self.binding_out = None
        self.destination = None
        self.req_info = None
        self.op_type = ""

    def verify_request(self, query, binding):
        """
        :param query: The SAML query, transport encoded
        :param binding: Which binding the query came in over
        """
        resp_args = {}
        if not query:
            logger.info("Missing QUERY")
            resp = Unauthorized('Unknown user')
            return resp_args, resp(self.environ, self.start_response)

        if not self.req_info:
            self.req_info = IDP.parse_authn_request(query, binding)

        logger.info("parsed OK")
        _authn_req = self.req_info.message
        logger.debug("%s", _authn_req)

        try:
            self.binding_out, self.destination = IDP.pick_binding(
                "assertion_consumer_service",
                bindings=self.response_bindings,
                entity_id=_authn_req.issuer.text)
        except Exception as err:
            logger.error("Couldn't find receiver endpoint: %s", err)
            raise

        logger.debug("Binding: %s, destination: %s", self.binding_out,
                                                       self.destination)

        resp_args = {}
        try:
            resp_args = IDP.response_args(_authn_req)
            _resp = None
        except UnknownPrincipal as excp:
            _resp = IDP.create_error_response(_authn_req.id,
                                              self.destination, excp)
        except UnsupportedBinding as excp:
            _resp = IDP.create_error_response(_authn_req.id,
                                              self.destination, excp)

        return resp_args, _resp

    def do(self, query, binding_in, relay_state="", encrypt_cert=None):
        """

        :param query: The request
        :param binding_in: Which binding was used when receiving the query
        :param relay_state: The relay state provided by the SP
        :param encrypt_cert: Cert to use for encryption
        :return: A response
        """
        try:
            resp_args, _resp = self.verify_request(query, binding_in)
        except UnknownPrincipal as excp:
            logger.error("UnknownPrincipal: %s", excp)
            resp = ServiceError("UnknownPrincipal: %s" % (excp,))
            return resp(self.environ, self.start_response)
        except UnsupportedBinding as excp:
            logger.error("UnsupportedBinding: %s", excp)
            resp = ServiceError("UnsupportedBinding: %s" % (excp,))
            return resp(self.environ, self.start_response)

        if not _resp:
            identity = USERS[self.user].copy()
            # identity["eduPersonTargetedID"] = get_eptid(IDP, query, session)
            logger.info("Identity: %s", identity)

            if REPOZE_ID_EQUIVALENT:
                identity[REPOZE_ID_EQUIVALENT] = self.user
            try:
                try:
                    metod = self.environ["idp.authn"]
                except KeyError:
                    pass
                else:
                    resp_args["authn"] = metod

                _resp = IDP.create_authn_response(
                    identity, userid=self.user,
                    encrypt_cert=encrypt_cert,
                    **resp_args)
            except Exception as excp:
                logging.error(exception_trace(excp))
                resp = ServiceError("Exception: %s" % (excp,))
                return resp(self.environ, self.start_response)

        logger.info("AuthNResponse: %s", _resp)
        if self.op_type == "ecp":
            kwargs = {"soap_headers": [
                ecp.Response(
                    assertion_consumer_service_url=self.destination)]}
        else:
            kwargs = {}

        http_args = IDP.apply_binding(self.binding_out,
                                      "%s" % _resp, self.destination,
                                      relay_state, response=True, **kwargs)

        logger.debug("HTTPargs: %s", http_args)
        return self.response(self.binding_out, http_args)

    def _store_request(self, saml_msg):
        logger.debug("_store_request: %s", saml_msg)
        key = sha1(saml_msg["SAMLRequest"]).hexdigest()
        # store the AuthnRequest
        IDP.ticket[key] = saml_msg
        return key

    def redirect(self):
        """ This is the HTTP-redirect endpoint """

        logger.info("--- In SSO Redirect ---")
        saml_msg = self.unpack_redirect()

        try:
            _key = saml_msg["key"]
            saml_msg = IDP.ticket[_key]
            self.req_info = saml_msg["req_info"]
            del IDP.ticket[_key]
        except KeyError:
            try:
                self.req_info = IDP.parse_authn_request(saml_msg["SAMLRequest"],
                                                        BINDING_HTTP_REDIRECT)
            except KeyError:
                resp = BadRequest("Message signature verification failure")
                return resp(self.environ, self.start_response)

            _req = self.req_info.message

            if "SigAlg" in saml_msg and "Signature" in saml_msg:  # Signed
            # request
                issuer = _req.issuer.text
                _certs = IDP.metadata.certs(issuer, "any", "signing")
                verified_ok = False
                for cert in _certs:
                    if verify_redirect_signature(saml_msg, IDP.sec.sec_backend,
                                                 cert):
                        verified_ok = True
                        break
                if not verified_ok:
                    resp = BadRequest("Message signature verification failure")
                    return resp(self.environ, self.start_response)

            if self.user:
                if _req.force_authn:
                    saml_msg["req_info"] = self.req_info
                    key = self._store_request(saml_msg)
                    return self.not_authn(key, _req.requested_authn_context)
                else:
                    return self.operation(saml_msg, BINDING_HTTP_REDIRECT)
            else:
                saml_msg["req_info"] = self.req_info
                key = self._store_request(saml_msg)
                return self.not_authn(key, _req.requested_authn_context)
        else:
            return self.operation(saml_msg, BINDING_HTTP_REDIRECT)

    def post(self):
        """
        The HTTP-Post endpoint
        """
        logger.info("--- In SSO POST ---")
        saml_msg = self.unpack_either()
        self.req_info = IDP.parse_authn_request(
            saml_msg["SAMLRequest"], BINDING_HTTP_POST)
        _req = self.req_info.message
        if self.user:
            if _req.force_authn:
                saml_msg["req_info"] = self.req_info
                key = self._store_request(saml_msg)
                return self.not_authn(key, _req.requested_authn_context)
            else:
                return self.operation(saml_msg, BINDING_HTTP_POST)
        else:
            saml_msg["req_info"] = self.req_info
            key = self._store_request(saml_msg)
            return self.not_authn(key, _req.requested_authn_context)

    # def artifact(self):
    # # Can be either by HTTP_Redirect or HTTP_POST
    #     _req = self._store_request(self.unpack_either())
    #     if isinstance(_req, basestring):
    #         return self.not_authn(_req)
    #     return self.artifact_operation(_req)

    def ecp(self):
        # The ECP interface
        logger.info("--- ECP SSO ---")
        resp = None

        try:
            authz_info = self.environ["HTTP_AUTHORIZATION"]
            if authz_info.startswith("Basic "):
                try:
                    _info = base64.b64decode(authz_info[6:])
                except TypeError:
                    resp = Unauthorized()
                else:
                    try:
                        (user, passwd) = _info.split(":")
                        if is_equal(PASSWD[user], passwd):
                            resp = Unauthorized()
                        self.user = user
                        self.environ[
                            "idp.authn"] = AUTHN_BROKER.get_authn_by_accr(
                            PASSWORD)
                    except ValueError:
                        resp = Unauthorized()
            else:
                resp = Unauthorized()
        except KeyError:
            resp = Unauthorized()

        if resp:
            return resp(self.environ, self.start_response)

        _dict = self.unpack_soap()
        self.response_bindings = [BINDING_PAOS]
        # Basic auth ?!
        self.op_type = "ecp"
        return self.operation(_dict, BINDING_SOAP)


# -----------------------------------------------------------------------------
# === Authentication ====
# -----------------------------------------------------------------------------


def do_authentication(environ, start_response, authn_context, key,
                      redirect_uri):
    """
    Display the login form
    """
    logger.debug("Do authentication")
    auth_info = AUTHN_BROKER.pick(authn_context)

    if len(auth_info):
        method, reference = auth_info[0]
        logger.debug("Authn chosen: %s (ref=%s)", method, reference)
        return method(environ, start_response, reference, key, redirect_uri)
    else:
        resp = Unauthorized("No usable authentication method")
        return resp(environ, start_response)


# -----------------------------------------------------------------------------

PASSWD = {
    "daev0001": "qwerty",
    "haho0032": "qwerty",
    "roland": "dianakra",
    "babs": "howes",
    "upper": "crust"}


def username_password_authn(environ, start_response, reference, key,
                            redirect_uri):
    """
    Display the login form
    """
    logger.info("The login page")
    headers = []

    resp = Response(mako_template="login.mako", template_lookup=LOOKUP,
                    headers=headers)

    argv = {
        "action": "/verify",
        "login": "",
        "password": "",
        "key": key,
        "authn_reference": reference,
        "redirect_uri": redirect_uri
    }
    logger.info("do_authentication argv: %s", argv)
    return resp(environ, start_response, **argv)


def verify_username_and_password(dic):
    global PASSWD
    # verify username and password
    if PASSWD[dic["login"][0]] == dic["password"][0]:
        return True, dic["login"][0]
    else:
        return False, ""


def do_verify(environ, start_response, _):
    query = parse_qs(get_post(environ))

    logger.debug("do_verify: %s", query)

    try:
        _ok, user = verify_username_and_password(query)
    except KeyError:
        _ok = False
        user = None

    if not _ok:
        resp = Unauthorized("Unknown user or wrong password")
    else:
        uid = rndstr(24)
        IDP.cache.uid2user[uid] = user
        IDP.cache.user2uid[user] = uid
        logger.debug("Register %s under '%s'", user, uid)

        kaka = set_cookie("idpauthn", "/", uid, query["authn_reference"][0])

        lox = "%s?id=%s&key=%s" % (query["redirect_uri"][0], uid,
                                   query["key"][0])
        logger.debug("Redirect => %s", lox)
        resp = Redirect(lox, headers=[kaka], content="text/html")

    return resp(environ, start_response)


def not_found(environ, start_response):
    """Called if no URL matches."""
    resp = NotFound()
    return resp(environ, start_response)


# -----------------------------------------------------------------------------
# === Single log out ===
# -----------------------------------------------------------------------------

# def _subject_sp_info(req_info):
#    # look for the subject
#    subject = req_info.subject_id()
#    subject = subject.text.strip()
#    sp_entity_id = req_info.message.issuer.text.strip()
#    return subject, sp_entity_id

class SLO(Service):
    def do(self, request, binding, relay_state="", encrypt_cert=None):
        logger.info("--- Single Log Out Service ---")
        try:
            _, body = request.split("\n")
            logger.debug("req: '%s'", body)
            req_info = IDP.parse_logout_request(body, binding)
        except Exception as exc:
            logger.error("Bad request: %s", exc)
            resp = BadRequest("%s" % exc)
            return resp(self.environ, self.start_response)

        msg = req_info.message
        if msg.name_id:
            lid = IDP.ident.find_local_id(msg.name_id)
            logger.info("local identifier: %s", lid)
            if lid in IDP.cache.user2uid:
                uid = IDP.cache.user2uid[lid]
                if uid in IDP.cache.uid2user:
                    del IDP.cache.uid2user[uid]
                del IDP.cache.user2uid[lid]
            # remove the authentication
            try:
                IDP.session_db.remove_authn_statements(msg.name_id)
            except KeyError as exc:
                logger.error("ServiceError: %s", exc)
                resp = ServiceError("%s" % exc)
                return resp(self.environ, self.start_response)

        resp = IDP.create_logout_response(msg, [binding])

        try:
            hinfo = IDP.apply_binding(binding, "%s" % resp, "", relay_state)
        except Exception as exc:
            logger.error("ServiceError: %s", exc)
            resp = ServiceError("%s" % exc)
            return resp(self.environ, self.start_response)

        #_tlh = dict2list_of_tuples(hinfo["headers"])
        delco = delete_cookie(self.environ, "idpauthn")
        if delco:
            hinfo["headers"].append(delco)
        logger.info("Header: %s", (hinfo["headers"],))
        resp = Response(hinfo["data"], headers=hinfo["headers"])
        return resp(self.environ, self.start_response)


# ----------------------------------------------------------------------------
# Manage Name ID service
# ----------------------------------------------------------------------------


class NMI(Service):
    def do(self, query, binding, relay_state="", encrypt_cert=None):
        logger.info("--- Manage Name ID Service ---")
        req = IDP.parse_manage_name_id_request(query, binding)
        request = req.message

        # Do the necessary stuff
        name_id = IDP.ident.handle_manage_name_id_request(
            request.name_id, request.new_id, request.new_encrypted_id,
            request.terminate)

        logger.debug("New NameID: %s", name_id)

        _resp = IDP.create_manage_name_id_response(request)

        # It's using SOAP binding
        hinfo = IDP.apply_binding(BINDING_SOAP, "%s" % _resp, "",
                                  relay_state, response=True)

        resp = Response(hinfo["data"], headers=hinfo["headers"])
        return resp(self.environ, self.start_response)


# ----------------------------------------------------------------------------
# === Assertion ID request ===
# ----------------------------------------------------------------------------


# Only URI binding
class AIDR(Service):
    def do(self, aid, binding, relay_state="", encrypt_cert=None):
        logger.info("--- Assertion ID Service ---")

        try:
            assertion = IDP.create_assertion_id_request_response(aid)
        except Unknown:
            resp = NotFound(aid)
            return resp(self.environ, self.start_response)

        hinfo = IDP.apply_binding(BINDING_URI, "%s" % assertion, response=True)

        logger.debug("HINFO: %s", hinfo)
        resp = Response(hinfo["data"], headers=hinfo["headers"])
        return resp(self.environ, self.start_response)

    def operation(self, _dict, binding, **kwargs):
        logger.debug("_operation: %s", _dict)
        if not _dict or "ID" not in _dict:
            resp = BadRequest('Error parsing request or no request')
            return resp(self.environ, self.start_response)

        return self.do(_dict["ID"], binding, **kwargs)


# ----------------------------------------------------------------------------
# === Artifact resolve service ===
# ----------------------------------------------------------------------------

class ARS(Service):
    def do(self, request, binding, relay_state="", encrypt_cert=None):
        _req = IDP.parse_artifact_resolve(request, binding)

        msg = IDP.create_artifact_response(_req, _req.artifact.text)

        hinfo = IDP.apply_binding(BINDING_SOAP, "%s" % msg, "", "",
                                  response=True)

        resp = Response(hinfo["data"], headers=hinfo["headers"])
        return resp(self.environ, self.start_response)


# ----------------------------------------------------------------------------
# === Authn query service ===
# ----------------------------------------------------------------------------


# Only SOAP binding
class AQS(Service):
    def do(self, request, binding, relay_state="", encrypt_cert=None):
        logger.info("--- Authn Query Service ---")
        _req = IDP.parse_authn_query(request, binding)
        _query = _req.message

        msg = IDP.create_authn_query_response(_query.subject,
                                              _query.requested_authn_context,
                                              _query.session_index)

        logger.debug("response: %s", msg)
        hinfo = IDP.apply_binding(BINDING_SOAP, "%s" % msg, "", "",
                                  response=True)

        resp = Response(hinfo["data"], headers=hinfo["headers"])
        return resp(self.environ, self.start_response)


# ----------------------------------------------------------------------------
# === Attribute query service ===
# ----------------------------------------------------------------------------


# Only SOAP binding
class ATTR(Service):
    def do(self, request, binding, relay_state="", encrypt_cert=None):
        logger.info("--- Attribute Query Service ---")

        _req = IDP.parse_attribute_query(request, binding)
        _query = _req.message

        name_id = _query.subject.name_id
        uid = name_id.text
        logger.debug("Local uid: %s", uid)
        identity = EXTRA[self.user]

        # Comes in over SOAP so only need to construct the response
        args = IDP.response_args(_query, [BINDING_SOAP])
        msg = IDP.create_attribute_response(identity,
                                            name_id=name_id, **args)

        logger.debug("response: %s", msg)
        hinfo = IDP.apply_binding(BINDING_SOAP, "%s" % msg, "", "",
                                  response=True)

        resp = Response(hinfo["data"], headers=hinfo["headers"])
        return resp(self.environ, self.start_response)


# ----------------------------------------------------------------------------
# Name ID Mapping service
# When an entity that shares an identifier for a principal with an identity
# provider wishes to obtain a name identifier for the same principal in a
# particular format or federation namespace, it can send a request to
# the identity provider using this protocol.
# ----------------------------------------------------------------------------


class NIM(Service):
    def do(self, query, binding, relay_state="", encrypt_cert=None):
        req = IDP.parse_name_id_mapping_request(query, binding)
        request = req.message
        # Do the necessary stuff
        try:
            name_id = IDP.ident.handle_name_id_mapping_request(
                request.name_id, request.name_id_policy)
        except Unknown:
            resp = BadRequest("Unknown entity")
            return resp(self.environ, self.start_response)
        except PolicyError:
            resp = BadRequest("Unknown entity")
            return resp(self.environ, self.start_response)

        info = IDP.response_args(request)
        _resp = IDP.create_name_id_mapping_response(name_id, **info)

        # Only SOAP
        hinfo = IDP.apply_binding(BINDING_SOAP, "%s" % _resp, "", "",
                                  response=True)

        resp = Response(hinfo["data"], headers=hinfo["headers"])
        return resp(self.environ, self.start_response)


# ----------------------------------------------------------------------------
# Cookie handling
# ----------------------------------------------------------------------------
def info_from_cookie(kaka):
    logger.debug("KAKA: %s", kaka)
    if kaka:
        cookie_obj = SimpleCookie(kaka)
        morsel = cookie_obj.get("idpauthn", None)
        if morsel:
            try:
                key, ref = base64.b64decode(morsel.value).split(":")
                return IDP.cache.uid2user[key], ref
            except (TypeError, KeyError):
                return None, None
        else:
            logger.debug("No idpauthn cookie")
    return None, None


def delete_cookie(environ, name):
    kaka = environ.get("HTTP_COOKIE", '')
    logger.debug("delete KAKA: %s", kaka)
    if kaka:
        cookie_obj = SimpleCookie(kaka)
        morsel = cookie_obj.get(name, None)
        cookie = SimpleCookie()
        cookie[name] = ""
        cookie[name]['path'] = "/"
        logger.debug("Expire: %s", morsel)
        cookie[name]["expires"] = _expiration("dawn")
        return tuple(cookie.output().split(": ", 1))
    return None


def set_cookie(name, _, *args):
    cookie = SimpleCookie()
    cookie[name] = base64.b64encode(":".join(args))
    cookie[name]['path'] = "/"
    cookie[name]["expires"] = _expiration(5)  # 5 minutes from now
    logger.debug("Cookie expires: %s", cookie[name]["expires"])
    return tuple(cookie.output().split(": ", 1))

# ----------------------------------------------------------------------------

# map urls to functions
AUTHN_URLS = [
    # sso
    (r'sso/post$', (SSO, "post")),
    (r'sso/post/(.*)$', (SSO, "post")),
    (r'sso/redirect$', (SSO, "redirect")),
    (r'sso/redirect/(.*)$', (SSO, "redirect")),
    (r'sso/art$', (SSO, "artifact")),
    (r'sso/art/(.*)$', (SSO, "artifact")),
    # slo
    (r'slo/redirect$', (SLO, "redirect")),
    (r'slo/redirect/(.*)$', (SLO, "redirect")),
    (r'slo/post$', (SLO, "post")),
    (r'slo/post/(.*)$', (SLO, "post")),
    (r'slo/soap$', (SLO, "soap")),
    (r'slo/soap/(.*)$', (SLO, "soap")),
    #
    (r'airs$', (AIDR, "uri")),
    (r'ars$', (ARS, "soap")),
    # mni
    (r'mni/post$', (NMI, "post")),
    (r'mni/post/(.*)$', (NMI, "post")),
    (r'mni/redirect$', (NMI, "redirect")),
    (r'mni/redirect/(.*)$', (NMI, "redirect")),
    (r'mni/art$', (NMI, "artifact")),
    (r'mni/art/(.*)$', (NMI, "artifact")),
    (r'mni/soap$', (NMI, "soap")),
    (r'mni/soap/(.*)$', (NMI, "soap")),
    # nim
    (r'nim$', (NIM, "soap")),
    (r'nim/(.*)$', (NIM, "soap")),
    #
    (r'aqs$', (AQS, "soap")),
    (r'attr$', (ATTR, "soap"))
]

NON_AUTHN_URLS = [
    #(r'login?(.*)$', do_authentication),
    (r'verify?(.*)$', do_verify),
    (r'sso/ecp$', (SSO, "ecp")),
]

# ----------------------------------------------------------------------------


def metadata(environ, start_response):
    try:
        path = args.path
        if path is None or len(path) == 0:
            path = os.path.dirname(os.path.abspath(__file__))
        if path[-1] != "/":
            path += "/"
        metadata = create_metadata_string(path + args.config, IDP.config,
                                          args.valid, args.cert, args.keyfile,
                                          args.id, args.name, args.sign)
        start_response('200 OK', [('Content-Type', "text/xml")])
        return metadata
    except Exception as ex:
        logger.error("An error occured while creating metadata:", ex.message)
        return not_found(environ, start_response)


def staticfile(environ, start_response):
    try:
        path = args.path
        if path is None or len(path) == 0:
            path = os.path.dirname(os.path.abspath(__file__))
        if path[-1] != "/":
            path += "/"
        path += environ.get('PATH_INFO', '').lstrip('/')
        path = os.path.realpath(path)
        if not path.startswith(args.path):
            resp = Unauthorized()
            return resp(environ, start_response)
        start_response('200 OK', [('Content-Type', "text/xml")])
        return open(path, 'r').read()
    except Exception as ex:
        logger.error("An error occured while creating metadata:", ex.message)
        return not_found(environ, start_response)


def application(environ, start_response):
    """
    The main WSGI application. Dispatch the current request to
    the functions from above and store the regular expression
    captures in the WSGI environment as  `myapp.url_args` so that
    the functions from above can access the url placeholders.

    If nothing matches, call the `not_found` function.

    :param environ: The HTTP application environment
    :param start_response: The application to run when the handling of the
        request is done
    :return: The response as a list of lines
    """

    path = environ.get('PATH_INFO', '').lstrip('/')

    if path == "metadata":
        return metadata(environ, start_response)

    kaka = environ.get("HTTP_COOKIE", None)
    logger.info("<application> PATH: %s", path)

    if kaka:
        logger.info("= KAKA =")
        user, authn_ref = info_from_cookie(kaka)
        if authn_ref:
            environ["idp.authn"] = AUTHN_BROKER[authn_ref]
    else:
        try:
            query = parse_qs(environ["QUERY_STRING"])
            logger.debug("QUERY: %s", query)
            user = IDP.cache.uid2user[query["id"][0]]
        except KeyError:
            user = None

    url_patterns = AUTHN_URLS
    if not user:
        logger.info("-- No USER --")
        # insert NON_AUTHN_URLS first in case there is no user
        url_patterns = NON_AUTHN_URLS + url_patterns

    for regex, callback in url_patterns:
        match = re.search(regex, path)
        if match is not None:
            try:
                environ['myapp.url_args'] = match.groups()[0]
            except IndexError:
                environ['myapp.url_args'] = path

            logger.debug("Callback: %s", callback)
            if isinstance(callback, tuple):
                cls = callback[0](environ, start_response, user)
                func = getattr(cls, callback[1])
                return func()
            return callback(environ, start_response, user)

    if re.search(r'static/.*', path) is not None:
        return staticfile(environ, start_response)
    return not_found(environ, start_response)

# ----------------------------------------------------------------------------

# allow uwsgi or gunicorn mount
# by moving some initialization out of __name__ == '__main__' section.
# uwsgi -s 0.0.0.0:8088 --protocol http --callable application --module idp

args = type('Config', (object,), {})
args.config = 'idp_conf'
args.mako_root = './'
args.path = None

import socket
from idp_user import USERS
from idp_user import EXTRA
from mako.lookup import TemplateLookup

AUTHN_BROKER = AuthnBroker()
AUTHN_BROKER.add(authn_context_class_ref(PASSWORD),
                 username_password_authn, 10,
                 "http://%s" % socket.gethostname())
AUTHN_BROKER.add(authn_context_class_ref(UNSPECIFIED),
                 "", 0, "http://%s" % socket.gethostname())
CONFIG = importlib.import_module(args.config)
IDP = server.Server(args.config, cache=Cache())
IDP.ticket = {}

# ----------------------------------------------------------------------------

if __name__ == '__main__':
    from wsgiref.simple_server import make_server

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', dest='path', help='Path to configuration file.')
    parser.add_argument('-v', dest='valid',
                        help="How long, in days, the metadata is valid from "
                             "the time of creation")
    parser.add_argument('-c', dest='cert', help='certificate')
    parser.add_argument('-i', dest='id',
                        help="The ID of the entities descriptor")
    parser.add_argument('-k', dest='keyfile',
                        help="A file with a key to sign the metadata with")
    parser.add_argument('-n', dest='name')
    parser.add_argument('-s', dest='sign', action='store_true',
                        help="sign the metadata")
    parser.add_argument('-m', dest='mako_root', default="./")
    parser.add_argument(dest="config")
    args = parser.parse_args()

    _rot = args.mako_root
    LOOKUP = TemplateLookup(directories=[_rot + 'templates', _rot + 'htdocs'],
                            module_directory=_rot + 'modules',
                            input_encoding='utf-8', output_encoding='utf-8')

    HOST = CONFIG.HOST
    PORT = CONFIG.PORT

    SRV = make_server(HOST, PORT, application)
    print("IdP listening on %s:%s" % (HOST, PORT))
    SRV.serve_forever()
else:
    _rot = args.mako_root
    LOOKUP = TemplateLookup(directories=[_rot + 'templates', _rot + 'htdocs'],
                            module_directory=_rot + 'modules',
                            input_encoding='utf-8', output_encoding='utf-8')



# See http://technet.microsoft.com/en-us/library/cc733065(v=ws.10).aspx
# and http://technet.microsoft.com/en-us/library/ee913589(v=ws.10).aspx
# for information regarding the default claim types supported by
# Microsoft ADFS v1.x.

MAP = {
    "identifier": "urn:oasis:names:tc:SAML:2.0:attrname-format:unspecified",
    "fro": {
        'http://schemas.xmlsoap.org/claims/commonname': 'commonName',
        'http://schemas.xmlsoap.org/claims/emailaddress': 'emailAddress',
        'http://schemas.xmlsoap.org/claims/group': 'group',
        'http://schemas.xmlsoap.org/claims/upn': 'upn',
        },
    "to": {
        'commonName': 'http://schemas.xmlsoap.org/claims/commonname',
        'emailAddress': 'http://schemas.xmlsoap.org/claims/emailaddress',
        'group': 'http://schemas.xmlsoap.org/claims/group',
        'upn': 'http://schemas.xmlsoap.org/claims/upn',
    }
}



#!/usr/bin/env python
from saml2.saml import NAME_FORMAT_URI

__author__ = 'rolandh'

import json

BASE = "http://localhost:8088"

metadata = open("idp_test/idp.xml").read()

info = {
    "entity_id": "%s/idp.xml" % BASE,
    "interaction": [
        {
            "matches": {
                "url": "%s/login" % BASE,
                "title": 'IDP test login'
            },
            "page-type": "login",
            "control": {
                "type": "form",
                "set": {"login": "roland", "password": "dianakra"}
            }
        },
        {
            "matches": {
                "url": "%s/sso/redirect" % BASE,
                "title": "SAML 2.0 POST"
            },
            "page-type": "other",
            "control": {
                "index": 0,
                "type": "form",
                "set": {}
            }
        },
        {
            "matches": {
                "url": "%s/sso/post" % BASE,
                "title": "SAML 2.0 POST"
            },
            "page-type": "other",
            "control": {
                "index": 0,
                "type": "form",
                "set": {}
            }
        },
        {
            "matches": {
                "url": "%s/slo/post" % BASE,
                "title": "SAML 2.0 POST"
            },
            "page-type": "other",
            "control": {
                "index": 0,
                "type": "form",
                "set": {}
            }
        }
    ],
    "metadata": metadata,
    "name_format": NAME_FORMAT_URI
}

print(json.dumps(info))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import os
import sys
from importlib import import_module

from saml2.s_utils import factory
from saml2.s_utils import do_ava
from saml2 import saml, ExtensionElement, NAMESPACE
from saml2 import extension_elements_to_elements
from saml2 import SAMLError
from saml2.saml import NAME_FORMAT_UNSPECIFIED, NAMEID_FORMAT_PERSISTENT, NameID

import logging
logger = logging.getLogger(__name__)


class UnknownNameFormat(SAMLError):
    pass


class ConverterError(SAMLError):
    pass


def load_maps(dirspec):
    """ load the attribute maps

    :param dirspec: a directory specification
    :return: a dictionary with the name of the map as key and the
        map as value. The map itself is a dictionary with two keys:
        "to" and "fro". The values for those keys are the actual mapping.
    """
    mapd = {}
    if dirspec not in sys.path:
        sys.path.insert(0, dirspec)

    for fil in os.listdir(dirspec):
        if fil.endswith(".py"):
            mod = import_module(fil[:-3])
            for item in _find_maps_in_module(mod):
                mapd[item["identifier"]] = item

    return mapd


def ac_factory(path=""):
    """Attribute Converter factory

    :param path: The path to a directory where the attribute maps are expected
        to reside.
    :return: A list of AttributeConverter instances
    """
    acs = []

    if path:
        if path not in sys.path:
            sys.path.insert(0, path)

        for fil in sorted(os.listdir(path)):
            if fil.endswith(".py"):
                mod = import_module(fil[:-3])
                acs.extend(_attribute_map_module_to_acs(mod))
    else:
        from saml2 import attributemaps

        for typ in attributemaps.__all__:
            mod = import_module(".%s" % typ, "saml2.attributemaps")
            acs.extend(_attribute_map_module_to_acs(mod))

    return acs


def _attribute_map_module_to_acs(module):
    """Scan an attribute map module and return any attribute maps defined

    :param: module: the python map module
    :type: types.ModuleType
    :return: a generator yielding AttributeConverter defintions
    :rtype: typing.Iterable[AttributeConverter]
    """
    for item in _find_maps_in_module(module):
        atco = AttributeConverter(item["identifier"])
        atco.from_dict(item)
        yield atco


def _find_maps_in_module(module):
    """Find attribute map dictionaries in a map file

    :param: module: the python map module
    :type: types.ModuleType
    :return: a generator yielding dict objects which have the right shape
    :rtype: typing.Iterable[dict]
    """
    for key, item in module.__dict__.items():
        if key.startswith("__"):
            continue
        if isinstance(item, dict) and "identifier" in item and (
            "to" in item or "fro" in item
        ):
            yield item


def to_local(acs, statement, allow_unknown_attributes=False):
    """ Replaces the attribute names in a attribute value assertion with the
    equivalent name from a local name format.

    :param acs: List of Attribute Converters
    :param statement: The Attribute Statement
    :param allow_unknown_attributes: If unknown attributes are allowed
    :return: A key,values dictionary
    """
    return list_to_local(acs, statement.attribute, allow_unknown_attributes)


def list_to_local(acs, attrlist, allow_unknown_attributes=False):
    """ Replaces the attribute names in a attribute value assertion with the
    equivalent name from a local name format.

    :param acs: List of Attribute Converters
    :param attrlist: List of Attributes
    :param allow_unknown_attributes: If unknown attributes are allowed
    :return: A key,values dictionary
    """
    if not acs:
        acs = [AttributeConverter()]
        acsd = {"": acs}
    else:
        acsd = dict([(a.name_format, a) for a in acs])

    ava = {}
    for attr in attrlist:
        try:
            _func = acsd[attr.name_format].ava_from
        except KeyError:
            if attr.name_format == NAME_FORMAT_UNSPECIFIED or \
                    allow_unknown_attributes:
                _func = acs[0].lcd_ava_from
            else:
                logger.info("Unsupported attribute name format: %s",
                    attr.name_format)
                continue

        try:
            key, val = _func(attr)
        except KeyError:
            if allow_unknown_attributes:
                key, val = acs[0].lcd_ava_from(attr)
            else:
                logger.info("Unknown attribute name: %s", attr)
                continue
        except AttributeError:
            continue

        try:
            ava[key].extend(val)
        except KeyError:
            ava[key] = val

    return ava


def from_local(acs, ava, name_format):
    for aconv in acs:
        #print(ac.format, name_format)
        if aconv.name_format == name_format:
            #print("Found a name_form converter")
            return aconv.to_(ava)

    return None


def from_local_name(acs, attr, name_format):
    """
    :param acs: List of AttributeConverter instances
    :param attr: attribute name as string
    :param name_format: Which name-format it should be translated to
    :return: An Attribute instance
    """
    for aconv in acs:
        #print(ac.format, name_format)
        if aconv.name_format == name_format:
            #print("Found a name_form converter")
            return aconv.to_format(attr)
    return attr


def to_local_name(acs, attr):
    """
    :param acs: List of AttributeConverter instances
    :param attr: an Attribute instance
    :return: The local attribute name
    """
    for aconv in acs:
        lattr = aconv.from_format(attr)
        if lattr:
            return lattr

    return attr.friendly_name


def get_local_name(acs, attr, name_format):
    for aconv in acs:
        #print(ac.format, name_format)
        if aconv.name_format == name_format:
            return aconv._fro.get(attr)


def d_to_local_name(acs, attr):
    """
    :param acs: List of AttributeConverter instances
    :param attr: an Attribute dictionary
    :return: The local attribute name
    """
    for aconv in acs:
        lattr = aconv.d_from_format(attr)
        if lattr:
            return lattr

    # if everything else fails this might be good enough
    try:
        return attr["friendly_name"]
    except KeyError:
        raise ConverterError("Could not find local name for %s" % attr)


class AttributeConverter(object):
    """ Converts from an attribute statement to a key,value dictionary and
        vice-versa """

    def __init__(self, name_format=""):
        self.name_format = name_format
        self._to = None
        self._fro = None

    def adjust(self):
        """ If one of the transformations is not defined it is expected to
        be the mirror image of the other.
        """

        if self._fro is None and self._to is not None:
            self._fro = dict(
                [(value.lower(), key) for key, value in self._to.items()])
        if self._to is None and self._fro is not None:
            self._to = dict(
                [(value.lower(), key) for key, value in self._fro.items()])

    def from_dict(self, mapdict):
        """ Import the attribute map from  a dictionary

        :param mapdict: The dictionary
        """

        self.name_format = mapdict["identifier"]
        try:
            self._fro = dict(
                [(k.lower(), v) for k, v in mapdict["fro"].items()])
        except KeyError:
            pass
        try:
            self._to = dict([(k.lower(), v) for k, v in mapdict["to"].items()])
        except KeyError:
            pass

        if self._fro is None and self._to is None:
            raise ConverterError("Missing specifications")

        if self._fro is None or self._to is None:
            self.adjust()

    def lcd_ava_from(self, attribute):
        """
        If nothing else works, this should

        :param attribute: an Attribute instance
        :return:
        """
        name = attribute.name.strip()
        values = [
            (value.text or '').strip()
            for value in attribute.attribute_value]
        return name, values

    def fail_safe_fro(self, statement):
        """ In case there is not formats defined or if the name format is
        undefined

        :param statement: AttributeStatement instance
        :return: A dictionary with names and values
        """
        result = {}
        for attribute in statement.attribute:
            if attribute.name_format and \
                    attribute.name_format != NAME_FORMAT_UNSPECIFIED:
                continue
            try:
                name = attribute.friendly_name.strip()
            except AttributeError:
                name = attribute.name.strip()

            result[name] = []
            for value in attribute.attribute_value:
                if not value.text:
                    result[name].append('')
                else:
                    result[name].append(value.text.strip())
        return result

    def ava_from(self, attribute, allow_unknown=False):
        try:
            attr = self._fro[attribute.name.strip().lower()]
        except AttributeError:
            attr = attribute.friendly_name.strip().lower()
        except KeyError:
            if allow_unknown:
                try:
                    attr = attribute.name.strip().lower()
                except AttributeError:
                    attr = attribute.friendly_name.strip().lower()
            else:
                raise

        val = []
        for value in attribute.attribute_value:
            if value.extension_elements:
                ext = extension_elements_to_elements(value.extension_elements,
                                                     [saml])
                for ex in ext:
                    if attr == "eduPersonTargetedID" and ex.text:
                        val.append(ex.text.strip())
                    else:
                        cval = {}
                        for key, (name, typ, mul) in ex.c_attributes.items():
                            exv = getattr(ex, name)
                            if exv:
                                cval[name] = exv
                        if ex.text:
                            cval["value"] = ex.text.strip()
                        val.append({ex.c_tag: cval})
            elif not value.text:
                val.append('')
            else:
                val.append(value.text.strip())

        return attr, val

    def fro(self, statement):
        """ Get the attributes and the attribute values.

        :param statement: The AttributeStatement.
        :return: A dictionary containing attributes and values
        """

        if not self.name_format:
            return self.fail_safe_fro(statement)

        result = {}
        for attribute in statement.attribute:
            if attribute.name_format and self.name_format and \
                    attribute.name_format != self.name_format:
                continue

            try:
                (key, val) = self.ava_from(attribute)
            except (KeyError, AttributeError):
                pass
            else:
                result[key] = val

        return result

    def to_format(self, attr):
        """ Creates an Attribute instance with name, name_format and
        friendly_name

        :param attr: The local name of the attribute
        :return: An Attribute instance
        """
        try:
            _attr = self._to[attr]
        except KeyError:
            try:
                _attr = self._to[attr.lower()]
            except:
                _attr = ''

        if _attr:
            return factory(saml.Attribute,
                           name=_attr,
                           name_format=self.name_format,
                           friendly_name=attr)
        else:
            return factory(saml.Attribute, name=attr)

    def from_format(self, attr):
        """ Find out the local name of an attribute

        :param attr: An saml.Attribute instance
        :return: The local attribute name or "" if no mapping could be made
        """
        if attr.name_format:
            if self.name_format == attr.name_format:
                try:
                    return self._fro[attr.name.lower()]
                except KeyError:
                    pass
        else:  # don't know the name format so try all I have
            try:
                return self._fro[attr.name.lower()]
            except KeyError:
                pass

        return ""

    def d_from_format(self, attr):
        """ Find out the local name of an attribute

        :param attr: An Attribute dictionary
        :return: The local attribute name or "" if no mapping could be made
        """
        if attr["name_format"]:
            if self.name_format == attr["name_format"]:
                try:
                    return self._fro[attr["name"].lower()]
                except KeyError:
                    pass
        else:  # don't know the name format so try all I have
            try:
                return self._fro[attr["name"].lower()]
            except KeyError:
                pass

        return ""

    def to_(self, attrvals):
        """ Create a list of Attribute instances.

        :param attrvals: A dictionary of attributes and values
        :return: A list of Attribute instances
        """
        attributes = []
        for key, value in attrvals.items():
            name = self._to.get(key.lower())
            if name:
                if name == "urn:oid:1.3.6.1.4.1.5923.1.1.1.10":
                    # special case for eduPersonTargetedID
                    attr_value = self.to_eptid_value(value)
                else:
                    attr_value = do_ava(value)
                attributes.append(factory(saml.Attribute,
                                          name=name,
                                          name_format=self.name_format,
                                          friendly_name=key,
                                          attribute_value=attr_value))
            else:
                attributes.append(factory(saml.Attribute,
                                          name=key,
                                          attribute_value=do_ava(value)))

        return attributes

    def to_eptid_value(self, values):
        """
        Create AttributeValue instances of NameID from the given values.

        Special handling for the "eptid" attribute
        Name=urn:oid:1.3.6.1.4.1.5923.1.1.1.10
        FriendlyName=eduPersonTargetedID

        values is a list of items of type str or dict. When an item is a
        dictionary it has the keys: "NameQualifier", "SPNameQualifier", and
        "text".

        Returns a list of AttributeValue instances of NameID elements.
        """

        def _create_nameid_ext_el(value):
            text = value["text"] if isinstance(value, dict) else value
            attributes = (
                {
                    "Format": NAMEID_FORMAT_PERSISTENT,
                    "NameQualifier": value["NameQualifier"],
                    "SPNameQualifier": value["SPNameQualifier"],
                }
                if isinstance(value, dict)
                else {"Format": NAMEID_FORMAT_PERSISTENT}
            )
            element = ExtensionElement(
                "NameID", NAMESPACE, attributes=attributes, text=text
            )
            return element

        attribute_values = [
            saml.AttributeValue(extension_elements=[_create_nameid_ext_el(v)])
            for v in values
        ]
        return attribute_values


class AttributeConverterNOOP(AttributeConverter):
    """ Does a NOOP conversion, that is no conversion is made """

    def __init__(self, name_format=""):
        AttributeConverter.__init__(self, name_format)

    def to_(self, attrvals):
        """ Create a list of Attribute instances.

        :param attrvals: A dictionary of attributes and values
        :return: A list of Attribute instances
        """
        attributes = []
        for key, value in attrvals.items():
            key = key.lower()
            attributes.append(factory(saml.Attribute,
                                      name=key,
                                      name_format=self.name_format,
                                      attribute_value=do_ava(value)))

        return attributes



from six.moves.urllib.parse import parse_qs
from saml2.authn_context import INTERNETPROTOCOLPASSWORD
from saml2.samlp import attribute_query_from_string, logout_request_from_string
from saml2 import BINDING_HTTP_REDIRECT, pack
from saml2 import BINDING_HTTP_POST
from saml2 import BINDING_SOAP
from saml2.server import Server
from saml2.soap import parse_soap_enveloped_saml_attribute_query
from saml2.soap import parse_soap_enveloped_saml_logout_request
from saml2.soap import make_soap_enveloped_saml_thingy

__author__ = 'rolandh'

TYP = {
    "GET": [BINDING_HTTP_REDIRECT],
    "POST": [BINDING_HTTP_POST, BINDING_SOAP]
}


AUTHN = {
    "class_ref": INTERNETPROTOCOLPASSWORD,
    "authn_auth": "http://www.example.com/login"
}


def unpack_form(_str, ver="SAMLRequest"):
    SR_STR = "name=\"%s\" value=\"" % ver
    RS_STR = 'name="RelayState" value="'

    i = _str.find(SR_STR)
    i += len(SR_STR)
    j = _str.find('"', i)

    sr = _str[i:j]

    k = _str.find(RS_STR, j)
    k += len(RS_STR)
    l = _str.find('"', k)

    rs = _str[k:l]

    return {ver: sr, "RelayState": rs}


class DummyResponse(object):
    def __init__(self, status, data, headers=None):
        self.status_code = status
        self.text = data
        self.headers = headers or []
        self.content = data


class FakeIDP(Server):
    def __init__(self, config_file=""):
        Server.__init__(self, config_file)
        #self.sign = False

    def receive(self, url, method="GET", **kwargs):
        """
        Interface to receive HTTP calls on

        :param url:
        :param method:
        :param kwargs:
        :return:
        """

        if method == "GET":
            path, query = url.split("?")
            qs_dict = parse_qs(kwargs["data"])
            req = qs_dict["SAMLRequest"][0]
            rstate = qs_dict["RelayState"][0]
        else:
            # Could be either POST or SOAP
            path = url
            try:
                qs_dict = parse_qs(kwargs["data"])
                req = qs_dict["SAMLRequest"][0]
                rstate = qs_dict["RelayState"][0]
            except KeyError:
                req = kwargs["data"]
                rstate = ""

        response = ""

        # Get service from path

        for key, vals in self.config.getattr("endpoints", "idp").items():
            for endp, binding in vals:
                if path == endp:
                    assert binding in TYP[method]
                    if key == "single_sign_on_service":
                        return self.authn_request_endpoint(req, binding,
                                                           rstate)
                    elif key == "single_logout_service":
                        return self.logout_endpoint(req, binding)

        for key, vals in self.config.getattr("endpoints", "aa").items():
            for endp, binding in vals:
                if path == endp:
                    assert binding in TYP[method]
                    if key == "attribute_service":
                        return self.attribute_query_endpoint(req, binding)

        return response

    def authn_request_endpoint(self, req, binding, relay_state):
        req = self.parse_authn_request(req, binding)
        if req.message.protocol_binding == BINDING_HTTP_REDIRECT:
            _binding = BINDING_HTTP_POST
        else:
            _binding = req.message.protocol_binding

        try:
            resp_args = self.response_args(req.message, [_binding])
        except Exception:
            raise

        identity = {"surName": "Hedberg", "givenName": "Roland",
                    "title": "supertramp", "mail": "roland@example.com"}
        userid = "Pavill"

        authn_resp = self.create_authn_response(identity,
                                                userid=userid,
                                                authn=AUTHN,
                                                **resp_args)

        response = "%s" % authn_resp

        _dict = pack.factory(_binding, response,
                             resp_args["destination"], relay_state,
                             "SAMLResponse")
        return DummyResponse(**_dict)

    def attribute_query_endpoint(self, xml_str, binding):
        if binding == BINDING_SOAP:
            _str = parse_soap_enveloped_saml_attribute_query(xml_str)
        else:
            _str = xml_str

        aquery = attribute_query_from_string(_str)
        extra = {"eduPersonAffiliation": "faculty"}
        #userid = "Pavill"

        name_id = aquery.subject.name_id
        attr_resp = self.create_attribute_response(
            extra, aquery.id, None, sp_entity_id=aquery.issuer.text,
            name_id=name_id, attributes=aquery.attribute)

        if binding == BINDING_SOAP:
            # SOAP packing
            #headers = {"content-type": "application/soap+xml"}
            soap_message = make_soap_enveloped_saml_thingy(attr_resp)
            #            if self.sign and self.sec:
            #                _signed = self.sec.sign_statement_using_xmlsec(soap_message,
            #                                                               class_name(attr_resp),
            #                                                               nodeid=attr_resp.id)
            #                soap_message = _signed
            response = "%s" % soap_message
        else:  # Just POST
            response = "%s" % attr_resp

        return DummyResponse(status=200, data=response)

    def logout_endpoint(self, xml_str, binding):
        if binding == BINDING_SOAP:
            _str = parse_soap_enveloped_saml_logout_request(xml_str)
        else:
            _str = xml_str

        req = logout_request_from_string(_str)

        _resp = self.create_logout_response(req, [binding])

        if binding == BINDING_SOAP:
            # SOAP packing
            #headers = {"content-type": "application/soap+xml"}
            soap_message = make_soap_enveloped_saml_thingy(_resp)
            #            if self.sign and self.sec:
            #                _signed = self.sec.sign_statement_using_xmlsec(soap_message,
            #                                                               class_name(attr_resp),
            #                                                               nodeid=attr_resp.id)
            #                soap_message = _signed
            response = "%s" % soap_message
        else: # Just POST
            response = "%s" % _resp

        return DummyResponse(status=200, data=response)



#!/usr/bin/env python
from Cookie import SimpleCookie
import logging
import os
import sp_conf
from sp_conf import CONFIG
import re
import subprocess
from urlparse import parse_qs
import argparse
from saml2 import BINDING_HTTP_REDIRECT, time_util
from saml2.httputil import Response
from saml2.httputil import Unauthorized
from saml2.httputil import NotFound
from saml2.httputil import Redirect
#from saml2.httputil import ServiceError
from saml2.metadata import create_metadata_string
from saml2.metadata import entities_descriptor
from saml2.config import Config
from saml2.sigver import security_context

logger = logging.getLogger("saml2.SP")

args = None
# -----------------------------------------------------------------------------


def dict_to_table(ava, lev=0, width=1):
    txt = ['<table border=%s bordercolor="black">\n' % width]
    for prop, valarr in ava.items():
        txt.append("<tr>\n")
        if isinstance(valarr, basestring):
            txt.append("<th>%s</th>\n" % str(prop))
            try:
                txt.append("<td>%s</td>\n" % valarr.encode("utf8"))
            except AttributeError:
                txt.append("<td>%s</td>\n" % valarr)
        elif isinstance(valarr, list):
            i = 0
            n = len(valarr)
            for val in valarr:
                if not i:
                    txt.append("<th rowspan=%d>%s</td>\n" % (len(valarr), prop))
                else:
                    txt.append("<tr>\n")
                if isinstance(val, dict):
                    txt.append("<td>\n")
                    txt.extend(dict_to_table(val, lev + 1, width - 1))
                    txt.append("</td>\n")
                else:
                    try:
                        txt.append("<td>%s</td>\n" % val.encode("utf8"))
                    except AttributeError:
                        txt.append("<td>%s</td>\n" % val)
                if n > 1:
                    txt.append("</tr>\n")
                n -= 1
                i += 1
        elif isinstance(valarr, dict):
            txt.append("<th>%s</th>\n" % prop)
            txt.append("<td>\n")
            txt.extend(dict_to_table(valarr, lev + 1, width - 1))
            txt.append("</td>\n")
        txt.append("</tr>\n")
    txt.append('</table>\n')
    return txt


def _expiration(timeout, tformat=None):
    if timeout == "now":
        return time_util.instant(tformat)
    else:
        # validity time should match lifetime of assertions
        return time_util.in_a_while(minutes=timeout, format=tformat)


def delete_cookie(environ, name):
    kaka = environ.get("HTTP_COOKIE", '')
    if kaka:
        cookie_obj = SimpleCookie(kaka)
        morsel = cookie_obj.get(name, None)
        cookie = SimpleCookie()
        cookie[name] = morsel
        cookie[name]["expires"] = _expiration("now",
                                              "%a, %d-%b-%Y %H:%M:%S CET")
        return tuple(cookie.output().split(": ", 1))
    return None

# ----------------------------------------------------------------------------


#noinspection PyUnusedLocal
def whoami(environ, start_response, user):
    nameid = environ["repoze.who.identity"]["login"]
    ava = environ["repoze.who.identity"]["user"]
    if not nameid:
        return not_authn(environ, start_response)
    if ava:
        response = ["<h2>Your identity is supposed to be</h2>"]
        response.extend(dict_to_table(ava))
    else:
        response = [
            "<h2>The system did not return any information about you</h2>"]

    response.extend("<a href='logout'>Logout</a>")
    resp = Response(response)
    return resp(environ, start_response)


#noinspection PyUnusedLocal
def not_found(environ, start_response):
    """Called if no URL matches."""
    resp = NotFound('Not Found')
    return resp(environ, start_response)


#noinspection PyUnusedLocal
def not_authn(environ, start_response):
    resp = Unauthorized('Unknown user')
    return resp(environ, start_response)


#noinspection PyUnusedLocal
def slo(environ, start_response, user):
    # so here I might get either a LogoutResponse or a LogoutRequest
    client = environ['repoze.who.plugins']["saml2auth"]
    sc = client.saml_client

    if "QUERY_STRING" in environ:
        query = parse_qs(environ["QUERY_STRING"])
        logger.info("query: %s", query)
        try:
            response = sc.parse_logout_request_response(
                query["SAMLResponse"][0], binding=BINDING_HTTP_REDIRECT)
            if response:
                logger.info("LOGOUT response parsed OK")
        except KeyError:
            # return error reply
            response = None

        if response is None:
            request = sc.lo

    headers = []
    delco = delete_cookie(environ, "pysaml2")
    if delco:
        headers.append(delco)
    resp = Redirect("/done", headers=headers)
    return resp(environ, start_response)


#noinspection PyUnusedLocal
def logout(environ, start_response, user):
    # This is where it starts when a user wants to log out
    client = environ['repoze.who.plugins']["saml2auth"]
    subject_id = environ["repoze.who.identity"]['repoze.who.userid']
    logger.info("[logout] subject_id: '%s'", subject_id)
    target = "/done"

    # What if more than one
    _dict = client.saml_client.global_logout(subject_id)
    logger.info("[logout] global_logout > %s", _dict)
    rem = environ['repoze.who.plugins'][client.rememberer_name]
    rem.forget(environ, subject_id)

    for key, item in _dict.items():
        if isinstance(item, tuple):
            binding, htargs = item
        else:  # result from logout, should be OK
            pass

    resp = Redirect("Successful Logout", headers=[("Location", target)])
    return resp(environ, start_response)
    # else:
    #     resp = ServiceError("Failed to logout from identity services")
    #     start_response("500 Internal Server Error")
    #     return []


#noinspection PyUnusedLocal
def done(environ, start_response, user):
    # remove cookie and stored info
    logger.info("[done] environ: %s", environ)
    subject_id = environ["repoze.who.identity"]['repoze.who.userid']
    client = environ['repoze.who.plugins']["saml2auth"]
    logger.info("[logout done] remaining subjects: %s",
        client.saml_client.users.subjects())

    start_response('200 OK', [('Content-Type', 'text/html')])
    return ["<h3>You are now logged out from this service</h3>"]

# ----------------------------------------------------------------------------

# map urls to functions
urls = [
    (r'whoami$', whoami),
    (r'logout$', logout),
    (r'done$', done),
    (r'slo$', slo),
    (r'^$', whoami),
]

# ----------------------------------------------------------------------------

def metadata(environ, start_response):
    try:
        path = args.path
        if path is None or len(path) == 0:
            path = os.path.dirname(os.path.abspath( __file__ ))
        if path[-1] != "/":
            path += "/"
        metadata = create_metadata_string(path+"sp_conf.py", None,
                                          args.valid, args.cert, args.keyfile,
                                          args.id, args.name, args.sign)
        start_response('200 OK', [('Content-Type', "text/xml")])
        return metadata
    except Exception as ex:
        logger.error("An error occured while creating metadata: %s", ex.message)
        return not_found(environ, start_response)

def application(environ, start_response):
    """
    The main WSGI application. Dispatch the current request to
    the functions from above and store the regular expression
    captures in the WSGI environment as `myapp.url_args` so that
    the functions from above can access the url placeholders.

    If nothing matches, call the `not_found` function.

    :param environ: The HTTP application environment
    :param start_response: The application to run when the handling of the
        request is done
    :return: The response as a list of lines
    """
    path = environ.get('PATH_INFO', '').lstrip('/')
    logger.info("<application> PATH: %s", path)

    if path == "metadata":
        return metadata(environ, start_response)

    user = environ.get("REMOTE_USER", "")
    if not user:
        user = environ.get("repoze.who.identity", "")
        logger.info("repoze.who.identity: '%s'", user)
    else:
        logger.info("REMOTE_USER: '%s'", user)
    #logger.info(logging.Logger.manager.loggerDict)
    for regex, callback in urls:
        if user:
            match = re.search(regex, path)
            if match is not None:
                try:
                    environ['myapp.url_args'] = match.groups()[0]
                except IndexError:
                    environ['myapp.url_args'] = path
                return callback(environ, start_response, user)
        else:
            return not_authn(environ, start_response)

    return not_found(environ, start_response)

# ----------------------------------------------------------------------------

from repoze.who.config import make_middleware_with_config

app_with_auth = make_middleware_with_config(application, {"here": "."},
                                            './who.ini',
                                            log_file="repoze_who.log")

# ----------------------------------------------------------------------------
HOST = sp_conf.HOST
PORT = sp_conf.PORT

# allow uwsgi or gunicorn mount
# by moving some initialization out of __name__ == '__main__' section.
# uwsgi -s 0.0.0.0:8087 --protocol http --callable app_with_auth --module idp

if __name__ == '__main__':
    #make_metadata arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', dest='path', help='Path to configuration file.')
    parser.add_argument('-v', dest='valid', default="4",
                        help="How long, in days, the metadata is valid from the time of creation")
    parser.add_argument('-c', dest='cert', help='certificate')
    parser.add_argument('-i', dest='id',
                        help="The ID of the entities descriptor in the metadata")
    parser.add_argument('-k', dest='keyfile',
                        help="A file with a key to sign the metadata with")
    parser.add_argument('-n', dest='name')
    parser.add_argument('-s', dest='sign', action='store_true',
                        help="sign the metadata")
    args = parser.parse_args()

    from wsgiref.simple_server import make_server
    srv = make_server(HOST, PORT, app_with_auth)
    print("SP listening on %s:%s" % (HOST, PORT))
    srv.serve_forever()



from contextlib import closing
from six.moves.urllib.parse import parse_qs
from six.moves.urllib.parse import urlparse
from saml2.authn_context import INTERNETPROTOCOLPASSWORD
from saml2.samlp import AuthnRequest
from saml2.samlp import NameIDPolicy
from saml2.saml import Assertion
from saml2.saml import NAMEID_FORMAT_TRANSIENT
from saml2 import BINDING_HTTP_POST
from saml2 import BINDING_URI
from saml2 import BINDING_SOAP
from saml2.client import Saml2Client
from saml2.server import Server

__author__ = 'rolandh'

TAG1 = "name=\"SAMLRequest\" value="


AUTHN = {
    "class_ref": INTERNETPROTOCOLPASSWORD,
    "authn_auth": "http://www.example.com/login"
}


def get_msg(hinfo, binding, response=False):
    if binding == BINDING_SOAP:
        msg = hinfo["data"]
    elif binding == BINDING_HTTP_POST:
        _inp = hinfo["data"]
        i = _inp.find(TAG1)
        i += len(TAG1) + 1
        j = _inp.find('"', i)
        msg = _inp[i:j]
    elif binding == BINDING_URI:
        if response:
            msg = hinfo["data"]
        else:
            msg = ""
            return parse_qs(hinfo["url"].split("?")[1])["ID"][0]
    else:  # BINDING_HTTP_REDIRECT
        parts = urlparse(hinfo["headers"][0][1])
        msg = parse_qs(parts.query)["SAMLRequest"][0]

    return msg


def test_basic_flow():
    sp = Saml2Client(config_file="servera_conf")
    with closing(Server(config_file="idp_all_conf")) as idp:
        # -------- @IDP -------------

        relay_state = "FOO"
        # -- dummy request ---
        orig_req = AuthnRequest(
            issuer=sp._issuer(), name_id_policy=NameIDPolicy(
                allow_create="true", format=NAMEID_FORMAT_TRANSIENT))

        # == Create an AuthnRequest response

        name_id = idp.ident.transient_nameid("id12", sp.config.entityid)

        binding, destination = idp.pick_binding("assertion_consumer_service",
                                                entity_id=sp.config.entityid)
        resp = idp.create_authn_response({"eduPersonEntitlement": "Short stop",
                                          "surName": "Jeter",
                                          "givenName": "Derek",
                                          "mail": "derek.jeter@nyy.mlb.com",
                                          "title": "The man"},
                                         "id-123456789",
                                         destination,
                                         sp.config.entityid,
                                         name_id=name_id,
                                         authn=AUTHN)

        hinfo = idp.apply_binding(binding, "%s" % resp, destination, relay_state)

        # --------- @SP -------------

        xmlstr = get_msg(hinfo, binding)
        # Explicitly allow unsigned responses for this test
        sp.want_response_signed = False
        aresp = sp.parse_authn_request_response(xmlstr, binding,
                                                {resp.in_response_to: "/"})

        # == Look for assertion X

        asid = aresp.assertion.id

        binding, destination = sp.pick_binding("assertion_id_request_service",
                                               entity_id=idp.config.entityid)

        hinfo = sp.apply_binding(binding, asid, destination)

        # ---------- @IDP ------------

        aid = get_msg(hinfo, binding, response=False)

        # == construct response

        resp = idp.create_assertion_id_request_response(aid)

        hinfo = idp.apply_binding(binding, "%s" % resp, None, "", response=True)

        # ----------- @SP -------------

        xmlstr = get_msg(hinfo, binding, response=True)

        final = sp.parse_assertion_id_request_response(xmlstr, binding)

        print(final.response)
        assert isinstance(final.response, Assertion)
